# Generated by Django 3.2.3 on 2022-07-30 09:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0005_customsoftdelete"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomTableName",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("int_field", models.IntegerField(null=True, unique=True)),
            ],
            options={
                "db_table": "order",
            },
        ),
    ]



# Generated by Django 5.2 on 2025-04-18 08:55

import django
from django.db import migrations, models


def get_operations():
    if django.VERSION >= (5, 2):
        return [
            migrations.CreateModel(
                name="SoftDeleteCompositePk",
                fields=[
                    ("id_1", models.IntegerField()),
                    ("id_2", models.IntegerField()),
                    (
                        "pk",
                        models.CompositePrimaryKey(
                            "id_1",
                            "id_2",
                            blank=True,
                            editable=False,
                            primary_key=True,
                            serialize=False,
                        ),
                    ),
                    ("is_active", models.BooleanField(default=True)),
                    ("other_field", models.TextField()),
                ],
            ),
        ]
    else:  # pragma: no cover
        return []


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0013_alter_testtrigger_m2m_field_changedcondition"),
    ]

    operations = get_operations()



# Generated by Django 3.0.7 on 2020-07-18 09:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0002_logentry_tologmodel"),
    ]

    operations = [
        migrations.AddField(
            model_name="logentry",
            name="new_field",
            field=models.CharField(max_length=16, null=True),
        ),
        migrations.AddField(
            model_name="logentry",
            name="old_field",
            field=models.CharField(max_length=16, null=True),
        ),
    ]



import contextlib
import logging

from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from pgtrigger import core, installation, registry, runtime


def _setup_logging():  # pragma: no cover
    installation.LOGGER.addHandler(logging.StreamHandler())
    if not installation.LOGGER.level:
        installation.LOGGER.setLevel(logging.INFO)


class SubCommands(BaseCommand):  # pragma: no cover
    """
    Subcommand class vendored in from
    https://github.com/andrewp-as-is/django-subcommands.py
    because of installation issues
    """

    argv = []
    subcommands = {}

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers(dest="subcommand", title="subcommands", description="")
        subparsers.required = True

        for command_name, command_class in self.subcommands.items():
            command = command_class()

            subparser = subparsers.add_parser(command_name, help=command_class.help)
            command.add_arguments(subparser)
            prog_name = subcommand = ""
            if self.argv:
                prog_name = self.argv[0]
                subcommand = self.argv[1]

            command_parser = command.create_parser(prog_name, subcommand)
            subparser._actions = command_parser._actions

    def run_from_argv(self, argv):
        self.argv = argv
        return super().run_from_argv(argv)

    def handle(self, *args, **options):
        command_name = options["subcommand"]
        self.subcommands.get(command_name)
        command_class = self.subcommands[command_name]

        if self.argv:
            args = [self.argv[0]] + self.argv[2:]
            return command_class().run_from_argv(args)
        else:
            return command_class().execute(*args, **options)


class BaseSchemaCommand(BaseCommand):
    """Sets the search path based on any "schema" option that's found"""

    def handle(self, *args, **options):
        database = options["database"] or DEFAULT_DB_ALIAS
        schemas = options["schema"] or []

        if schemas:
            context = runtime.schema(*schemas, databases=[database])
        else:
            context = contextlib.nullcontext()

        with context:
            return self.handle_with_schema(*args, **options)


class LsCommand(BaseSchemaCommand):
    help = "List triggers and their installation state."

    def add_arguments(self, parser):
        parser.add_argument("uris", nargs="*", type=str)
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )

    def handle_with_schema(self, *args, **options):
        uris = options["uris"]

        status_formatted = {
            core.UNINSTALLED: "\033[91mUNINSTALLED\033[0m",
            core.INSTALLED: "\033[92mINSTALLED\033[0m",
            core.OUTDATED: "\033[93mOUTDATED\033[0m",
            core.PRUNE: "\033[96mPRUNE\033[0m",
            core.UNALLOWED: "\033[94mUNALLOWED\033[0m",
        }

        enabled_formatted = {
            True: "\033[92mENABLED\033[0m",
            False: "\033[91mDISABLED\033[0m",
            None: "\033[94mN/A\033[0m",
        }

        def _format_status(status, enabled, uri):
            if status in (core.UNINSTALLED, core.UNALLOWED):
                enabled = None

            return status_formatted[status], enabled_formatted[enabled], uri

        formatted = []

        for model, trigger in registry.registered(*uris):
            uri = trigger.get_uri(model)
            status, enabled = trigger.get_installation_status(model, database=options["database"])
            formatted.append(_format_status(status, enabled, uri))

        if not uris:
            for trigger in installation.prunable(database=options["database"]):
                formatted.append(_format_status("PRUNE", trigger[2], f"{trigger[0]}:{trigger[1]}"))

        max_status_len = max(len(val) for val, _, _ in formatted)
        max_enabled_len = max(len(val) for _, val, _ in formatted)
        for status, enabled, uri in formatted:
            print(
                f"{{: <{max_status_len}}} {{: <{max_enabled_len}}} {{}}".format(
                    status, enabled, uri
                )
            )


class InstallCommand(BaseSchemaCommand):
    help = "Install triggers."

    def add_arguments(self, parser):
        parser.add_argument("uris", nargs="*", type=str)
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )

    def handle_with_schema(self, *args, **options):
        _setup_logging()
        installation.install(*options["uris"], database=options["database"])


class UninstallCommand(BaseSchemaCommand):
    help = "Uninstall triggers."

    def add_arguments(self, parser):
        parser.add_argument("uris", nargs="*", type=str)
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )

    def handle_with_schema(self, *args, **options):
        _setup_logging()
        installation.uninstall(*options["uris"], database=options["database"])


class EnableCommand(BaseSchemaCommand):
    help = "Enable triggers."

    def add_arguments(self, parser):
        parser.add_argument("uris", nargs="*", type=str)
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )

    def handle_with_schema(self, *args, **options):
        _setup_logging()
        installation.enable(*options["uris"], database=options["database"])


class DisableCommand(BaseSchemaCommand):
    help = "Disable triggers."

    def add_arguments(self, parser):
        parser.add_argument("uris", nargs="*", type=str)
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )

    def handle_with_schema(self, *args, **options):
        _setup_logging()
        installation.disable(*options["uris"], database=options["database"])


class PruneCommand(BaseSchemaCommand):
    help = "Prune installed triggers that are no longer in the codebase."

    def add_arguments(self, parser):
        parser.add_argument("-d", "--database", help="The database")
        parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )

    def handle_with_schema(self, *args, **options):
        _setup_logging()
        installation.prune(database=options["database"])


class Command(SubCommands):
    help = "Core django-pgtrigger subcommands."

    subcommands = {
        "ls": LsCommand,
        "install": InstallCommand,
        "uninstall": UninstallCommand,
        "enable": EnableCommand,
        "disable": DisableCommand,
        "prune": PruneCommand,
    }



"""Tests behavior related to migrations"""

import pathlib
import shutil
import time

import ddf
import django.contrib.auth.models as auth_models
import pytest
from django.apps import apps
from django.core.management import call_command
from django.db import models
from django.db.utils import ProgrammingError

import pgtrigger
import pgtrigger.tests.models as test_models
from pgtrigger import core
from pgtrigger.tests import utils


class SomeMixin:
    """For testing proxy model inheritance"""

    pass


@pytest.fixture(autouse=True)
def disble_install_on_migrate(settings):
    settings.PGTRIGGER_INSTALL_ON_MIGRATE = False


def migration_dir():
    return pathlib.Path(__file__).parent / "migrations"


def num_migration_files():
    return len(list(migration_dir().glob("0*.py")))


@pytest.fixture
def reset_triggers():
    """Ensures all triggers are uninstalled before the tests"""
    pgtrigger.uninstall(database="default")

    yield

    pgtrigger.install(database="default")


@pytest.fixture
def reset_migrations(tmp_path, request):
    """Ensures the migration dir is reset after the test"""
    num_orig_migrations = num_migration_files()
    shutil.copytree(migration_dir(), tmp_path / "migrations")

    try:
        yield
    finally:
        # Migrate back to the initial migration of the test to allevitate
        # some of the issues when re-using a test DB
        call_command(
            "migrate",
            "tests",
            str(num_orig_migrations).rjust(4, "0"),
            verbosity=request.config.option.verbose,
        )

        shutil.rmtree(migration_dir())
        shutil.copytree(tmp_path / "migrations", migration_dir())


def assert_all_triggers_installed():
    for model, trigger in pgtrigger.registered():
        status = trigger.get_installation_status(model)
        assert status[0] == core.INSTALLED


@pytest.mark.django_db(
    databases=["default", "other", "receipt", "order", "sqlite"], transaction=True
)
@pytest.mark.usefixtures("reset_triggers", "reset_migrations")
@pytest.mark.order(-1)  # This is a possibly leaky test if it fails midway. Always run last
def test_makemigrations_existing_models(settings, request):
    """
    Create migrations for existing models and test various scenarios
    where triggers are dynamically added and removed
    """
    # Verify that we've configured our test settings properly
    assert not settings.PGTRIGGER_INSTALL_ON_MIGRATE
    assert settings.PGTRIGGER_MIGRATIONS

    num_orig_migrations = num_migration_files()

    call_command("makemigrations", verbosity=request.config.option.verbose)
    assert num_migration_files() == num_orig_migrations + 1

    call_command("migrate", verbosity=request.config.option.verbose)
    assert_all_triggers_installed()

    # Add a new trigger to the registry that should be migrated
    trigger = pgtrigger.Trigger(
        when=pgtrigger.Before,
        name="my_migrated_trigger",
        operation=pgtrigger.Insert | pgtrigger.Update,
        func="RAISE EXCEPTION 'no no no!';",
    )

    with trigger.register(test_models.TestModel):
        call_command("makemigrations", verbosity=request.config.option.verbose)
        assert num_migration_files() == num_orig_migrations + 2

        # As a sanity check, ensure makemigrations doesnt make dups
        call_command("makemigrations", verbosity=request.config.option.verbose)
        assert num_migration_files() == num_orig_migrations + 2

        # Before migrating, I should be able to make a `TestModel`
        ddf.G("tests.TestModel")

        call_command("migrate", verbosity=request.config.option.verbose)
        assert_all_triggers_installed()

        # After migrating, test models should be protected
        with utils.raises_trigger_error(match="no no no!"):
            test_models.TestModel.objects.create()

        # Update the trigger to allow inserts, but not updates.
        # We should have a new migration
        trigger.operation = pgtrigger.Update
        call_command("makemigrations", verbosity=request.config.option.verbose)
        assert num_migration_files() == num_orig_migrations + 3

        call_command("migrate", verbosity=request.config.option.verbose)
        assert_all_triggers_installed()

        # We should be able to make test models but not update them
        test_model = ddf.G("tests.TestModel")
        with utils.raises_trigger_error(match="no no no!"):
            test_model.save()

    # The trigger is now removed from the registry. It should create
    # a new migration
    call_command("makemigrations", verbosity=request.config.option.verbose)
    assert num_migration_files() == num_orig_migrations + 4

    call_command("migrate", verbosity=request.config.option.verbose)
    assert_all_triggers_installed()

    # We should be able to create and update the test model now that
    # the trigger is removed
    test_model = ddf.G("tests.TestModel")
    test_model.save()

    # Create a protection trigger on the external user model and
    # migrate it
    trigger = pgtrigger.Trigger(
        when=pgtrigger.Before,
        name="nothing_allowed",
        operation=pgtrigger.Insert | pgtrigger.Update,
        func="RAISE EXCEPTION 'no no no!';",
    )

    # Test that special characters migrate correctly
    trigger = pgtrigger.Protect(
        name="special_characters",
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(new__char_field="%"),
    )
    with trigger.register(test_models.TestModel):
        call_command("makemigrations", verbosity=request.config.option.verbose)
        assert num_migration_files() == num_orig_migrations + 5

        call_command("migrate", verbosity=request.config.option.verbose)
        assert_all_triggers_installed()

        tm = ddf.G("tests.TestModel", char_field="hello")

        with utils.raises_trigger_error(match="Cannot update rows"):
            tm.char_field = "%"
            tm.save()


def make_migrations(atomic: bool):
    """Call makemigrations. Set atomic property of last migration if it is specified"""
    call_command("makemigrations", name=f"a{time.time()}".replace(".", ""))

    last_migration = sorted(pathlib.Path(migration_dir()).glob("[0-9]*.py"))[-1]
    with open(last_migration, "r") as f:
        contents = f.read()

    contents = contents.replace(
        "class Migration(migrations.Migration):\n",
        f"class Migration(migrations.Migration):\n    atomic = {atomic}\n",
    )
    with open(last_migration, "w") as f:
        f.write(contents)


@pytest.mark.django_db(
    databases=["default", "other", "receipt", "order", "sqlite"], transaction=True
)
@pytest.mark.usefixtures("reset_triggers", "reset_migrations")
@pytest.mark.order(-1)  # This is a possibly leaky test if it fails midway. Always run last
# Run independently of core test suite since since this creates/removes models
@pytest.mark.independent
@pytest.mark.parametrize("atomic", [True, False])
def test_makemigrations_create_remove_models(settings, atomic):
    """
    Tests migration scenarios where models are dynamically added and
    removed.
    """
    assert not settings.PGTRIGGER_INSTALL_ON_MIGRATE
    assert settings.PGTRIGGER_MIGRATIONS

    num_orig_migrations = num_migration_files()
    num_expected_migrations = num_orig_migrations

    ###
    # Make the initial trigger migrations
    ###
    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    ###
    # Create a new model, migrate it, and verify triggers
    ###
    class BaseDynamicTestModel(models.Model):
        field = models.CharField(max_length=120)
        user = models.ForeignKey(auth_models.User, on_delete=models.CASCADE)

        class Meta:
            abstract = True
            triggers = [
                pgtrigger.Protect(
                    name="protect_deletes",
                    operation=pgtrigger.Delete,
                    condition=~pgtrigger.Q(old__field="nothing"),
                ),
                pgtrigger.Protect(
                    name="protect_updates",
                    operation=pgtrigger.Update,
                    condition=~pgtrigger.Q(old__field="nothing"),
                ),
            ]

    class DynamicTestModel(BaseDynamicTestModel):
        pass

    test_models.DynamicTestModel = DynamicTestModel

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # Sanity check that we cannot delete or update a DynamicTestModel
    protected_model = ddf.G(test_models.DynamicTestModel)

    with utils.raises_trigger_error(match="Cannot update"):
        protected_model.field = "hello_world"
        protected_model.save()

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.delete()

    ###
    # Alter the column type when a condition depends on it. This should
    # correctly drop the trigger, update the column type, and add
    # the trigger
    ###
    class DynamicTestModel(BaseDynamicTestModel):
        field = models.TextField()

    test_models.DynamicTestModel = DynamicTestModel

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # Sanity check that we cannot delete or update a DynamicTestModel
    protected_model = ddf.G(test_models.DynamicTestModel)

    with utils.raises_trigger_error(match="Cannot update"):
        protected_model.field = "hello_world"
        protected_model.save()

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.delete()

    ###
    # Keep only deletion protection, migrate, and verify it's removed
    ###
    DynamicTestModel._meta.triggers = [
        pgtrigger.Protect(name="protect_deletes", operation=pgtrigger.Delete)
    ]
    DynamicTestModel._meta.original_attrs["triggers"] = DynamicTestModel._meta.triggers

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # Updates work, but deletes dont
    protected_model.field = "hello_there"
    protected_model.save()

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.delete()

    ###
    # Remove the model and verify it migrates
    ###
    del test_models.DynamicTestModel
    del apps.app_configs["tests"].models["dynamictestmodel"]
    apps.clear_cache()

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    with pytest.raises(ProgrammingError):
        protected_model.delete()

    # Unapply migration where a model with a trigger is removed
    # Any triggers that were defined on the model when it was removed should be
    # recreated.
    call_command("migrate", "tests", str(num_expected_migrations - 1).rjust(4, "0"))

    test_models.DynamicTestModel = DynamicTestModel
    protected_model = ddf.G(test_models.DynamicTestModel)
    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.delete()
    del test_models.DynamicTestModel

    # Reapply the migration we just unapplied
    call_command("migrate")

    # Create a new proxy model on a third-party app and add it to the test models
    class DynamicProxyModel(auth_models.User):
        class Meta:
            proxy = True
            triggers = [
                pgtrigger.Protect(name="protect_deletes", operation=pgtrigger.Delete),
                pgtrigger.Protect(name="protect_updates", operation=pgtrigger.Update),
            ]

    test_models.DynamicProxyModel = DynamicProxyModel

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # Sanity check that we cannot delete or update a user
    protected_model = ddf.G(auth_models.User)

    with utils.raises_trigger_error(match="Cannot update"):
        protected_model.username = "wes"
        protected_model.save()

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.delete()

    ###
    # Keep only deletion protection for proxy models and migrate
    ###
    DynamicProxyModel._meta.triggers = [
        pgtrigger.Protect(name="protect_deletes", operation=pgtrigger.Delete)
    ]
    DynamicProxyModel._meta.original_attrs["triggers"] = DynamicProxyModel._meta.triggers

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # Updates work, but deletes dont
    protected_model.username = "wes"
    protected_model.save()

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.delete()

    ###
    # Remove the proxy model and verify it migrates
    ###
    del test_models.DynamicProxyModel
    del apps.app_configs["tests"].models["dynamicproxymodel"]
    apps.clear_cache()

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # We can delete the original model
    protected_model.delete()

    # Create a new proxy model on auth_models.User group relationships and add it
    # to the test models
    class DynamicThroughModel(auth_models.User.groups.through):
        class Meta:
            proxy = True
            triggers = [
                pgtrigger.Protect(name="protect_deletes", operation=pgtrigger.Delete),
                pgtrigger.Protect(name="protect_inserts", operation=pgtrigger.Insert),
            ]

    test_models.DynamicThroughModel = DynamicThroughModel

    # Sanity check that we cannot insert or delete a group
    protected_model = ddf.G(auth_models.User)
    protected_model.groups.add(ddf.G(auth_models.Group))

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    with utils.raises_trigger_error(match="Cannot insert"):
        protected_model.groups.add(ddf.G(auth_models.Group))

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.groups.clear()

    ###
    # Proxy model inheritance
    # https://github.com/Opus10/django-pgtrigger/issues/126
    ###

    # Set up multiple inheritance for a proxy model with a mixin
    class SomeBaseModel(models.Model):
        pass

    class SomeProxyModel(SomeMixin, SomeBaseModel):
        class Meta:
            proxy = True

    test_models.SomeBaseModel = SomeBaseModel
    test_models.SomeProxyModel = SomeProxyModel

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    class SomeAbstractModel(models.Model):
        class Meta:
            abstract = True

    class SomeAbstractProxyModel(SomeAbstractModel, SomeBaseModel):
        class Meta:
            proxy = True

    test_models.SomeAbstractModel = SomeAbstractModel
    test_models.SomeAbstractProxyModel = SomeAbstractProxyModel

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    del test_models.SomeAbstractModel
    del test_models.SomeBaseModel
    del test_models.SomeProxyModel
    del test_models.SomeAbstractProxyModel
    del apps.app_configs["tests"].models["somebasemodel"]
    del apps.app_configs["tests"].models["someproxymodel"]
    del apps.app_configs["tests"].models["someabstractproxymodel"]
    apps.clear_cache()

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    ###
    # Keep only deletion protection for a dynamic through model and migrate
    ###
    DynamicThroughModel._meta.triggers = [
        pgtrigger.Protect(name="protect_deletes", operation=pgtrigger.Delete)
    ]
    DynamicThroughModel._meta.original_attrs["triggers"] = DynamicThroughModel._meta.triggers

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # Inserts work, but deletes dont
    protected_model.groups.add(ddf.G(auth_models.Group))

    with utils.raises_trigger_error(match="Cannot delete"):
        protected_model.groups.clear()

    # Remove the model and verify it migrates
    del test_models.DynamicThroughModel
    del apps.app_configs["tests"].models["dynamicthroughmodel"]
    apps.clear_cache()

    make_migrations(atomic)
    num_expected_migrations += 1
    assert num_migration_files() == num_expected_migrations
    call_command("migrate")
    assert_all_triggers_installed()

    # We can delete the groups
    protected_model.groups.clear()

    # Django has a known issue with using a default through model as a base in
    # migrations. We revert the migrations we just made up until the through model
    # so that the test doesn't pass when it cleans up all migrations
    call_command("migrate", "tests", str(num_orig_migrations + 8).rjust(4, "0"))



"""Additional goodies"""

import functools
import itertools
import operator
from typing import Any, List, Tuple, Union

from django.db import models

from pgtrigger import core, utils

# A sentinel value to determine if a kwarg is unset
_unset = object()


def _get_column(model, field):
    field = field if isinstance(field, models.Field) else model._meta.get_field(field)
    col = field.column
    if not col:  # pragma: no cover
        if getattr(field, "columns", None):
            raise ValueError(
                f"Field {field.name} references a composite key and is not supported."
            )
        else:
            raise ValueError(f"Field {field.name} does not reference a database column.")
    return col


def _get_columns(model, field):
    field = field if isinstance(field, models.Field) else model._meta.get_field(field)
    col = field.column
    cols = [col] if col else getattr(field, "columns", None)
    if not cols:  # pragma: no cover
        raise ValueError(f"Field {field.name} does not reference a database column.")
    return cols


class Protect(core.Trigger):
    """A trigger that raises an exception."""

    when: core.When = core.Before

    def get_func(self, model):
        sql = f"""
            RAISE EXCEPTION
                'pgtrigger: Cannot {str(self.operation).lower()} rows from % table',
                TG_TABLE_NAME;
        """
        return self.format_sql(sql)


class ReadOnly(Protect):
    """A trigger that prevents edits to fields.

    If `fields` are provided, will protect edits to only those fields.
    If `exclude` is provided, will protect all fields except the ones
    excluded.
    If none of these arguments are provided, all fields cannot be edited.
    """

    fields: Union[List[str], None] = None
    exclude: Union[List[str], None] = None
    operation: core.Operation = core.Update

    def __init__(
        self,
        *,
        fields: Union[List[str], None] = None,
        exclude: Union[List[str], None] = None,
        **kwargs: Any,
    ):
        self.fields = fields or self.fields
        self.exclude = exclude or self.exclude

        if self.fields and self.exclude:
            raise ValueError('Must provide only one of "fields" or "exclude" to ReadOnly trigger')

        super().__init__(**kwargs)

    def get_condition(self, model):
        if not self.fields and not self.exclude:
            return core.Condition("OLD.* IS DISTINCT FROM NEW.*")
        else:
            if self.exclude:
                # Sanity check that the exclude list contains valid fields
                for field in self.exclude:
                    model._meta.get_field(field)

                fields = [f.name for f in model._meta.fields if f.name not in self.exclude]
            else:
                fields = [model._meta.get_field(field).name for field in self.fields]

            return functools.reduce(
                operator.or_,
                [core.Q(**{f"old__{field}__df": core.F(f"new__{field}")}) for field in fields],
            )


class FSM(core.Trigger):
    """Enforces a finite state machine on a field.

    Supply the trigger with the `field` that transitions and then
    a list of tuples of valid transitions to the `transitions` argument.

    !!! note

        Only non-null `CharField` fields without quotes are currently supported.
        If your strings have a colon symbol in them, you must override the
        "separator" argument to be a value other than a colon.
    """

    when: core.When = core.Before
    operation: core.Operation = core.Update
    field: str = None
    transitions: List[Tuple[str, str]] = None
    separator: str = ":"

    def __init__(
        self,
        *,
        name: str = None,
        condition: Union[core.Condition, None] = None,
        field: str = None,
        transitions: List[Tuple[str, str]] = None,
        separator: str = None,
    ):
        self.field = field or self.field
        self.transitions = transitions or self.transitions
        self.separator = separator or self.separator

        if not self.field:  # pragma: no cover
            raise ValueError('Must provide "field" for FSM')

        if not self.transitions:  # pragma: no cover
            raise ValueError('Must provide "transitions" for FSM')

        # This trigger doesn't accept quoted values or values that
        # contain the configured separator
        for value in itertools.chain(*self.transitions):
            if "'" in value or '"' in value:
                raise ValueError(f'FSM transition value "{value}" contains quotes')
            elif self.separator in value:
                raise ValueError(
                    f'FSM value "{value}" contains separator "{self.separator}".'
                    ' Configure your trigger with a different "separator" attribute'
                )

        # The separator must be a single character that isn't a quote
        if len(self.separator) != 1:
            raise ValueError(f'Separator "{self.separator}" must be a single character')
        elif self.separator in ('"', "'"):
            raise ValueError("Separator must not have quotes")

        super().__init__(name=name, condition=condition)

    def get_declare(self, model):
        return [("_is_valid_transition", "BOOLEAN")]

    def get_func(self, model):
        col = _get_column(model, self.field)
        transition_uris = (
            "{" + ",".join([f"{old}{self.separator}{new}" for old, new in self.transitions]) + "}"
        )

        sql = f"""
            SELECT CONCAT(OLD.{utils.quote(col)}, '{self.separator}', NEW.{utils.quote(col)}) = ANY('{transition_uris}'::text[])
                INTO _is_valid_transition;

            IF (_is_valid_transition IS FALSE AND OLD.{utils.quote(col)} IS DISTINCT FROM NEW.{utils.quote(col)}) THEN
                RAISE EXCEPTION
                    'pgtrigger: Invalid transition of field "{self.field}" from "%" to "%" on table %',
                    OLD.{utils.quote(col)},
                    NEW.{utils.quote(col)},
                    TG_TABLE_NAME;
            ELSE
                RETURN NEW;
            END IF;
        """  # noqa
        return self.format_sql(sql)


class SoftDelete(core.Trigger):
    """Sets a field to a value when a delete happens.

    Supply the trigger with the "field" that will be set
    upon deletion and the "value" to which it should be set.
    The "value" defaults to `False`.

    !!! note

        This trigger currently only supports nullable `BooleanField`,
        `CharField`, and `IntField` fields.
    """

    when: core.When = core.Before
    operation: core.Operation = core.Delete
    field: str = None
    value: Union[bool, str, int, None] = False

    def __init__(
        self,
        *,
        name: str = None,
        condition: Union[core.Condition, None] = None,
        field: str = None,
        value: Union[bool, str, int, None] = _unset,
    ):
        self.field = field or self.field
        self.value = value if value is not _unset else self.value

        if not self.field:  # pragma: no cover
            raise ValueError('Must provide "field" for soft delete')

        super().__init__(name=name, condition=condition)

    def get_func(self, model):
        soft_field = _get_column(model, self.field)

        # Support composite primary keys in Django 5.2+
        pk_cols = _get_columns(model, model._meta.pk)
        table_pk_cols = ",".join(utils.quote(col) for col in pk_cols)
        trigger_pk_cols = ",".join(f"OLD.{utils.quote(col)}" for col in pk_cols)

        if len(pk_cols) > 1:
            table_pk_cols = f"({table_pk_cols})"
            trigger_pk_cols = f"({trigger_pk_cols})"

        def _render_value():
            if self.value is None:
                return "NULL"
            elif isinstance(self.value, str):
                return f"'{self.value}'"
            else:
                return str(self.value)

        sql = f"""
            UPDATE {utils.quote(model._meta.db_table)}
            SET {soft_field} = {_render_value()}
            WHERE {table_pk_cols} = {trigger_pk_cols};
            RETURN NULL;
        """
        return self.format_sql(sql)


class UpdateSearchVector(core.Trigger):
    """Updates a `django.contrib.postgres.search.SearchVectorField` from document fields.

    Supply the trigger with the `vector_field` that will be updated with
    changes to the `document_fields`. Optionally provide a `config_name`, which
    defaults to `pg_catalog.english`.

    This trigger uses `tsvector_update_trigger` to update the vector field.
    See [the Postgres docs](https://www.postgresql.org/docs/current/textsearch-features.html#TEXTSEARCH-UPDATE-TRIGGERS)
    for more information.

    !!! note

        `UpdateSearchVector` triggers are not compatible with [pgtrigger.ignore][] since
        it references a built-in trigger. Trying to ignore this trigger results in a
        `RuntimeError`.
    """  # noqa

    when: core.When = core.Before
    vector_field: str = None
    document_fields: List[str] = None
    config_name: str = "pg_catalog.english"

    def __init__(
        self,
        *,
        name: str = None,
        vector_field: str = None,
        document_fields: List[str] = None,
        config_name: str = None,
    ):
        self.vector_field = vector_field or self.vector_field
        self.document_fields = document_fields or self.document_fields
        self.config_name = config_name or self.config_name

        if not self.vector_field:
            raise ValueError('Must provide "vector_field" to update search vector')

        if not self.document_fields:
            raise ValueError('Must provide "document_fields" to update search vector')

        if not self.config_name:  # pragma: no cover
            raise ValueError('Must provide "config_name" to update search vector')

        super().__init__(name=name, operation=core.Insert | core.UpdateOf(*document_fields))

    def ignore(self, model):
        raise RuntimeError(f"Cannot ignore {self.__class__.__name__} triggers")

    def get_func(self, model):
        return ""

    def render_execute(self, model):
        document_cols = [_get_column(model, field) for field in self.document_fields]
        rendered_document_cols = ", ".join(utils.quote(col) for col in document_cols)
        vector_col = _get_column(model, self.vector_field)
        return (
            f"tsvector_update_trigger({utils.quote(vector_col)},"
            f" {utils.quote(self.config_name)}, {rendered_document_cols})"
        )



import copy

import dj_database_url

SECRET_KEY = "django-pgtrigger"
# Install the tests as an app so that we can make test models
INSTALLED_APPS = [
    "pgtrigger",
    # For testing purposes
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.postgres",
    "psqlextra",
    "pgtrigger.tests",
    "pgtrigger.tests.syncdb_app",
]

# Database url comes from the DATABASE_URL env var
# We have some multi-database and multi-schema tests
DATABASES = {
    "default": dj_database_url.config(),
    "sqlite": {"ENGINE": "django.db.backends.sqlite3", "NAME": "test_sqlite"},
}

DATABASES["other"] = copy.deepcopy(DATABASES["default"])
DATABASES["other"]["NAME"] += "_other"

DATABASES["default"]["ENGINE"] = "psqlextra.backend"

DATABASES["order"] = copy.deepcopy(DATABASES["default"])
DATABASES["order"]["OPTIONS"] = {"options": "-c search_path=order"}
DATABASES["receipt"] = copy.deepcopy(DATABASES["default"])
DATABASES["receipt"]["OPTIONS"] = {"options": "-c search_path=receipt"}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# Ensure partitioned models dont get migrated for non-default DBs
DATABASE_ROUTERS = ["pgtrigger.tests.models.Router"]

# Turn off pgtrigger migrations for normal manage.py use
PGTRIGGER_MIGRATIONS = False

# Ensure that we always install triggers if running locally
PGTRIGGER_INSTALL_ON_MIGRATE = True

USE_TZ = False



import django
from django.contrib.auth.models import User
from django.contrib.postgres.search import SearchVectorField
from django.db import connections, models
from django.utils import timezone
from psqlextra.models import PostgresPartitionedModel
from psqlextra.types import PostgresPartitioningMethod

import pgtrigger
import pgtrigger.utils


def _get_pg_maj_version(db):  # pragma: no cover
    connection = connections[db]
    if connection.vendor == "postgresql":
        with connection.cursor() as cursor:
            return pgtrigger.utils.pg_maj_version(cursor)


class Router:
    route_app_labels = ["tests"]

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        """
        Ignore the parititon model for the "other" DB, for non-Postgres DBs,
        and for Postgres DBs that are less than version 13
        """
        pg_maj_version = _get_pg_maj_version(db)

        if model_name == "partitionmodel" and (
            db in ("sqlite", "other") or not pg_maj_version or pg_maj_version < 13
        ):
            return False


class PartitionModel(PostgresPartitionedModel):
    class PartitioningMeta:
        method = PostgresPartitioningMethod.RANGE
        key = ["timestamp"]

    name = models.TextField()
    timestamp = models.DateTimeField()

    class Meta:
        triggers = [pgtrigger.Protect(name="protect_delete", operation=pgtrigger.Delete)]


class OrderSchema(models.Model):
    """A model that only appears in the "schema1" schema"""

    int_field = models.IntegerField()


class ReceiptSchema(models.Model):
    """A model that only appears in the "schema2" schema"""

    char_field = models.CharField(max_length=128)

    class Meta:
        db_table = "table.with.dots"


class SearchModel(models.Model):
    body_vector = SearchVectorField()
    title_body_vector = SearchVectorField()

    title = models.CharField(max_length=128)
    body = models.TextField()

    class Meta:
        triggers = [
            pgtrigger.UpdateSearchVector(
                name="add_body_to_vector", vector_field="body_vector", document_fields=["body"]
            ),
            pgtrigger.UpdateSearchVector(
                name="add_body_title_to_vector",
                vector_field="title_body_vector",
                document_fields=["body", "title"],
            ),
        ]


@pgtrigger.register(
    pgtrigger.Protect(name="protect_delete", operation=pgtrigger.Delete),
)
class CustomTableName(models.Model):
    int_field = models.IntegerField(null=True, unique=True)

    class Meta:
        db_table = "order"


class TestModel(models.Model):
    int_field = models.IntegerField(null=True, unique=True)
    char_field = models.CharField(max_length=128, null=True)
    float_field = models.FloatField(null=True)

    class Meta:
        unique_together = ("int_field", "char_field")


class LogEntry(models.Model):
    """Created when ToLogModel is updated"""

    level = models.CharField(max_length=16)
    old_field = models.CharField(max_length=16, null=True)
    new_field = models.CharField(max_length=16, null=True)


class ToLogModel(models.Model):
    """For testing triggers that log records at statement and row level"""

    field = models.CharField(max_length=16)

    class Meta:
        triggers = [
            pgtrigger.Trigger(
                name="update_of_statement_test",
                level=pgtrigger.Statement,
                operation=pgtrigger.UpdateOf("field"),
                when=pgtrigger.After,
                func=pgtrigger.Func(
                    f"""
                    INSERT INTO {LogEntry._meta.db_table}(level)
                    VALUES ('STATEMENT');
                    RETURN NULL;
                """
                ),
            ),
            pgtrigger.Trigger(
                name="after_update_statement_test",
                level=pgtrigger.Statement,
                operation=pgtrigger.Update,
                when=pgtrigger.After,
                referencing=pgtrigger.Referencing(old="old_values", new="new_values"),
                func=f"""
                    INSERT INTO {LogEntry._meta.db_table}(level, old_field, new_field)
                    SELECT 'STATEMENT' AS level,
                           old_values.field AS old_field,
                           new_values.field AS new_field
                         FROM old_values
                         JOIN new_values ON old_values.id = new_values.id;
                    RETURN NULL;
                """,
            ),
            pgtrigger.Trigger(
                name="after_update_row_test",
                level=pgtrigger.Row,
                operation=pgtrigger.Update,
                when=pgtrigger.After,
                condition=pgtrigger.Q(old__field__df=pgtrigger.F("new__field")),
                func=(
                    f"INSERT INTO {LogEntry._meta.db_table}(level) VALUES ('ROW'); RETURN NULL;"
                ),
            ),
        ]


class CharPk(models.Model):
    custom_pk = models.CharField(primary_key=True, max_length=32)


class TestTrigger(models.Model):
    """
    For testing triggers
    """

    field = models.CharField(max_length=16)
    int_field = models.IntegerField(default=0)
    dt_field = models.DateTimeField(default=timezone.now)
    nullable = models.CharField(null=True, default=None, max_length=16)
    fk_field = models.ForeignKey("auth.User", null=True, on_delete=models.CASCADE)
    char_pk_fk_field = models.ForeignKey(CharPk, null=True, on_delete=models.CASCADE)
    m2m_field = models.ManyToManyField(User, related_name="+")

    class Meta:
        triggers = [
            pgtrigger.Trigger(
                name="protect_misc_insert",
                when=pgtrigger.Before,
                operation=pgtrigger.Insert,
                func="RAISE EXCEPTION 'no no no!';",
                condition=pgtrigger.Q(new__field="misc_insert"),
            ),
        ]


class TestTriggerProxy(TestTrigger):
    """
    For testing triggers on proxy models
    """

    class Meta:
        proxy = True
        triggers = [
            pgtrigger.Protect(name="protect_delete", operation=pgtrigger.Delete),
        ]


class TestDefaultThrough(TestTrigger.m2m_field.through):
    class Meta:
        proxy = True
        triggers = [
            pgtrigger.Protect(name="protect_it", operation=pgtrigger.Delete),
        ]


@pgtrigger.register(pgtrigger.SoftDelete(name="soft_delete", field="is_active"))
class SoftDelete(models.Model):
    """
    For testing soft deletion. Deletions on this model will set
    is_active = False without deleting the model
    """

    is_active = models.BooleanField(default=True)
    other_field = models.TextField()


class FkToSoftDelete(models.Model):
    """Ensures foreign keys to a soft delete model are deleted"""

    ref = models.ForeignKey(SoftDelete, on_delete=models.CASCADE)


if django.VERSION >= (5, 2):

    @pgtrigger.register(pgtrigger.SoftDelete(name="soft_delete_composite_pk", field="is_active"))
    class SoftDeleteCompositePk(models.Model):
        """
        For testing soft deletion with a composite primary key.
        """

        id_1 = models.IntegerField()
        id_2 = models.IntegerField()
        pk = models.CompositePrimaryKey("id_1", "id_2")
        is_active = models.BooleanField(default=True)
        other_field = models.TextField()


@pgtrigger.register(pgtrigger.SoftDelete(name="soft_delete", field="custom_active"))
class CustomSoftDelete(models.Model):
    """
    For testing soft deletion with a custom active field.

    This trigger also helps ensure that triggers can have the same names
    across multiple models.
    """

    custom_active = models.BooleanField(default=True)
    other_field = models.TextField()


@pgtrigger.register(
    pgtrigger.FSM(
        name="fsm",
        field="transition",
        transitions=[("unpublished", "published"), ("published", "inactive")],
    )
)
class FSM(models.Model):
    """Tests valid transitions of a field"""

    transition = models.CharField(max_length=32)


class ChangedCondition(models.Model):
    """
    For testing changed conditions
    """

    field = models.CharField(max_length=16)
    int_field = models.IntegerField(default=0)
    dt_field = models.DateTimeField(auto_now=True)
    nullable = models.CharField(null=True, default=None, max_length=16)
    fk_field = models.ForeignKey("auth.User", null=True, on_delete=models.CASCADE)
    char_pk_fk_field = models.ForeignKey(CharPk, null=True, on_delete=models.CASCADE)
    m2m_field = models.ManyToManyField(User, related_name="+")



import ddf
import django
import pytest
from django.core.exceptions import FieldDoesNotExist

import pgtrigger
from pgtrigger.tests import models, utils


def test_registered_invalid_args():
    with pytest.raises(ValueError):
        pgtrigger.registered("uri")


@pytest.mark.django_db
def test_read_only():
    """Tests the ReadOnly trigger"""
    with pytest.raises(ValueError, match="only one of"):
        pgtrigger.ReadOnly(name="uneditable", fields=["level"], exclude=["hello"])

    trigger = pgtrigger.ReadOnly(name="uneditable", fields=["level"])
    with pytest.raises(FieldDoesNotExist):
        trigger.install(models.TestModel)

    trigger = pgtrigger.ReadOnly(name="uneditable", exclude=["level"])
    with pytest.raises(FieldDoesNotExist):
        trigger.install(models.TestModel)

    trigger = pgtrigger.ReadOnly(name="uneditable")
    with trigger.install(models.TestModel):
        m = ddf.G(models.TestModel, int_field=1)
        m.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            m.int_field = 2
            m.save()

    trigger = pgtrigger.ReadOnly(name="uneditable", fields=["char_field", "float_field"])
    with trigger.install(models.TestModel):
        m = ddf.G(models.TestModel, int_field=2, char_field="a")
        m.int_field = 3
        m.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            m.char_field = "b"
            m.save()

    trigger = pgtrigger.ReadOnly(name="uneditable", exclude=["int_field"])
    with trigger.install(models.TestModel):
        m = ddf.G(models.TestModel, int_field=4, char_field="a")
        m.int_field = 5
        m.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            m.char_field = "b"
            m.save()


@pytest.mark.django_db
def test_search_model():
    """Verifies search model fields are kept up to date"""
    obj = models.SearchModel.objects.create(
        title="This is a message", body="Hello World. What a great body."
    )
    models.SearchModel.objects.create(title="Hi guys", body="Random Word. This is a good idea.")
    models.SearchModel.objects.create(
        title="Hello", body="Other words. Many great ideas come from stuff."
    )
    models.SearchModel.objects.create(title="The title", body="A short message.")

    assert models.SearchModel.objects.filter(body_vector="hello").count() == 1
    assert models.SearchModel.objects.filter(body_vector="words").count() == 2
    assert models.SearchModel.objects.filter(body_vector="world").count() == 1
    assert models.SearchModel.objects.filter(title_body_vector="message").count() == 2
    assert models.SearchModel.objects.filter(title_body_vector="idea").count() == 2
    assert models.SearchModel.objects.filter(title_body_vector="hello").count() == 2

    obj.body = "Nothing more"
    obj.save()
    assert not models.SearchModel.objects.filter(body_vector="hello").exists()
    assert models.SearchModel.objects.filter(title_body_vector="hello").count() == 1


def test_update_search_vector_args():
    """Verifies arg checking for UpdateSearchVector"""
    with pytest.raises(ValueError, match='provide "vector_field"'):
        pgtrigger.UpdateSearchVector()

    with pytest.raises(ValueError, match='provide "document_fields"'):
        pgtrigger.UpdateSearchVector(vector_field="vector_field")


def test_update_search_vector_ignore():
    """Verifies UpdateSearchVector cannot be ignored"""
    trigger = pgtrigger.UpdateSearchVector(
        name="hi", vector_field="vector_field", document_fields=["hi"]
    )
    with pytest.raises(RuntimeError, match="Cannot ignore UpdateSearchVector"):
        with trigger.ignore(models.SearchModel):
            pass


@pytest.mark.django_db
def test_soft_delete():
    """
    Verifies the SoftDelete test model has the "is_active" flag set to false
    """
    soft_delete = ddf.G(models.SoftDelete, is_active=True)
    ddf.G(models.FkToSoftDelete, ref=soft_delete)
    soft_delete.delete()

    assert not models.SoftDelete.objects.get().is_active
    assert not models.FkToSoftDelete.objects.exists()


@pytest.mark.django_db
def test_soft_delete_composite_pk():
    """
    Verifies the SoftDelete test model has the "is_active" flag set to false
    """
    if django.VERSION >= (5, 2):
        models.SoftDeleteCompositePk.objects.create(is_active=True, id_1=1, id_2=1)
        soft_delete = models.SoftDeleteCompositePk.objects.create(is_active=True, id_1=2, id_2=2)
        soft_delete.delete()

        assert models.SoftDeleteCompositePk.objects.get(id_1=1, id_2=1).is_active
        assert not models.SoftDeleteCompositePk.objects.get(id_1=2, id_2=2).is_active


@pytest.mark.django_db
def test_customer_soft_delete():
    """
    Verifies the CustomSoftDelete test model has the "custom_active" flag set
    to false
    """
    soft_delete = ddf.G(models.CustomSoftDelete, custom_active=True)
    soft_delete.delete()

    assert not models.CustomSoftDelete.objects.get().custom_active


@pytest.mark.django_db
def test_soft_delete_different_values():
    """
    Tests SoftDelete with different types of fields and values
    """
    # Make the LogEntry model a soft delete model where
    # "level" is set to "inactive"
    trigger = pgtrigger.SoftDelete(name="soft_delete", field="level", value="inactive")
    with trigger.install(models.LogEntry):
        le = ddf.G(models.LogEntry, level="active")
        le.delete()
        assert models.LogEntry.objects.get().level == "inactive"
    models.LogEntry.objects.all().delete()

    # Make the LogEntry model a soft delete model where
    # "old_field" is set to None
    trigger = pgtrigger.SoftDelete(name="soft_delete", field="old_field", value=None)
    with trigger.install(models.LogEntry):
        le = ddf.G(models.LogEntry, old_field="something")
        le.delete()
        assert models.LogEntry.objects.get().old_field is None


@pytest.mark.django_db
def test_fsm():
    """
    Verifies the FSM test model cannot make invalid transitions
    """
    fsm = ddf.G(models.FSM, transition="unpublished")
    fsm.transition = "inactive"
    with utils.raises_trigger_error(match="Invalid transition"):
        fsm.save()

    fsm.transition = "published"
    fsm.save()

    # Be sure we ignore FSM when there is no transition
    fsm.save()

    with utils.raises_trigger_error(match="Invalid transition"):
        fsm.transition = "unpublished"
        fsm.save()

    fsm.transition = "inactive"
    fsm.save()


def test_fsm_args():
    """Verifies arg checking for FSM"""
    with pytest.raises(ValueError, match='provide "field"'):
        pgtrigger.FSM()

    with pytest.raises(ValueError, match='provide "transitions"'):
        pgtrigger.FSM(field="hello")

    with pytest.raises(ValueError, match='contains separator ":"'):
        pgtrigger.FSM(field="hello", transitions=[("a", ":")])

    with pytest.raises(ValueError, match='contains separator ","'):
        pgtrigger.FSM(field="hello", separator=",", transitions=[("a", ",")])

    with pytest.raises(ValueError, match="contains quotes"):
        pgtrigger.FSM(field="hello", transitions=[("a", "b'")])

    with pytest.raises(ValueError, match="contains quotes"):
        pgtrigger.FSM(field="hello", transitions=[("a", 'b"')])

    with pytest.raises(ValueError, match="single character"):
        pgtrigger.FSM(field="hello", separator="aa", transitions=[("a", "b")])

    with pytest.raises(ValueError, match="must not have quotes"):
        pgtrigger.FSM(field="hello", separator="'", transitions=[("a", "b")])


@pytest.mark.django_db
def test_protect():
    """Verify deletion protect trigger works on test model"""
    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()


@pytest.mark.django_db
def test_custom_db_table_protect_trigger():
    """Verify custom DB table names have successful triggers"""
    deletion_protected_model = ddf.G(models.CustomTableName)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, connections
from django.utils.version import get_version_tuple


def _psycopg_version():
    try:
        import psycopg as Database
    except ImportError:
        import psycopg2 as Database
    except Exception as exc:  # pragma: no cover
        raise ImproperlyConfigured("Error loading psycopg2 or psycopg module") from exc

    version_tuple = get_version_tuple(Database.__version__.split(" ", 1)[0])

    if version_tuple[0] not in (2, 3):  # pragma: no cover
        raise ImproperlyConfigured(f"Pysocpg version {version_tuple[0]} not supported")

    return version_tuple


psycopg_version = _psycopg_version()
psycopg_maj_version = psycopg_version[0]


class AttrDict(dict):
    """A dictionary where keys can be accessed as attributes"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


def connection(database=None):
    """
    Obtains the connection used for a trigger / model pair. The database
    for the connection is selected based on the write DB in the database
    router config.
    """
    return connections[database or DEFAULT_DB_ALIAS]


def pg_maj_version(cursor):
    """Return the major version of Postgres that's running"""
    version = getattr(cursor.connection, "server_version", cursor.connection.info.server_version)
    return int(str(version)[:-4])


def is_postgres(database):
    return connection(database).vendor == "postgresql"


def postgres_databases(databases=None):
    """Return postgres databases from the provided list of databases.

    If no databases are provided, return all postgres databases
    """
    databases = databases or list(settings.DATABASES)
    assert isinstance(databases, list)
    return [database for database in databases if is_postgres(database)]


def exec_sql(sql, database=None, fetchall=False):
    if is_postgres(database):  # pragma: no branch
        with connection(database).cursor() as cursor:
            cursor.execute(sql)

            if fetchall:
                return cursor.fetchall()


def quote(label, char='"'):
    """Conditionally wraps a label in quotes"""
    if label.startswith(char) or label.endswith(char):
        return label
    else:
        return f"{char}{label}{char}"


def render_uninstall(table, trigger_pgid):
    """Renders uninstallation SQL"""
    return f"DROP TRIGGER IF EXISTS {trigger_pgid} ON {quote(table)};"



import contextlib
import re
from collections.abc import Sequence
from typing import Any, Union

from django import __version__ as DJANGO_VERSION
from django.apps import apps
from django.db import transaction
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.operations.base import Operation
from django.db.migrations.operations.fields import AddField
from django.db.migrations.operations.models import CreateModel, IndexOperation
from django.db.migrations.state import ModelState, ProjectState
from django.db.models import Model

from pgtrigger import compiler, core, utils

if DJANGO_VERSION >= "5.1":
    from django.db.migrations.autodetector import OperationDependency

    OPERATION_DEPENDENCY_CREATE = OperationDependency.Type.CREATE
else:
    OperationDependency = lambda *args: tuple((args))  # noqa: E731
    OPERATION_DEPENDENCY_CREATE = True


def _add_trigger(
    schema_editor: BaseDatabaseSchemaEditor,
    model: type[Model],
    trigger: Union[core.Trigger, compiler.Trigger],
) -> None:
    """Add a trigger to a model."""
    if not isinstance(trigger, compiler.Trigger):  # pragma: no cover
        trigger = trigger.compile(model)

    with transaction.atomic(using=schema_editor.connection.alias):
        # Trigger install SQL returns interpolated SQL which makes
        # params=None a necessity to avoid escaping attempts on execution.
        schema_editor.execute(trigger.install_sql, params=None)


def _remove_trigger(
    schema_editor: BaseDatabaseSchemaEditor,
    model: type[Model],
    trigger: Union[core.Trigger, compiler.Trigger],
) -> None:
    """Remove a trigger from a model."""
    if not isinstance(trigger, compiler.Trigger):  # pragma: no cover
        trigger = trigger.compile(model)

    # Trigger uninstall SQL returns interpolated SQL which makes
    # params=None a necessity to avoid escaping attempts on execution.
    schema_editor.execute(trigger.uninstall_sql, params=None)


class TriggerOperationMixin:
    def allow_migrate_model_trigger(
        self, schema_editor: BaseDatabaseSchemaEditor, model: type[Model]
    ) -> bool:
        """
        The check for determinig if a trigger is migrated
        """
        assert isinstance(self, Operation)
        assert model._meta.concrete_model is not None
        return schema_editor.connection.vendor == "postgresql" and self.allow_migrate_model(
            schema_editor.connection.alias, model._meta.concrete_model
        )


class AddTrigger(TriggerOperationMixin, IndexOperation):
    option_name = "triggers"

    def __init__(self, model_name: str, trigger: Union[core.Trigger, compiler.Trigger]) -> None:
        self.model_name = model_name
        self.trigger = trigger

    def state_forwards(self, app_label: str, state: ProjectState) -> None:
        model_state = state.models[app_label, self.model_name]
        model_state.options["triggers"] = model_state.options.get("triggers", []) + [self.trigger]
        state.reload_model(app_label, self.model_name, delay=True)

    def database_forwards(
        self,
        app_label: str,
        schema_editor: BaseDatabaseSchemaEditor,
        from_state: ProjectState,
        to_state: ProjectState,
    ) -> None:
        model = to_state.apps.get_model(app_label, self.model_name)
        assert issubclass(model, Model)
        if self.allow_migrate_model_trigger(schema_editor, model):  # pragma: no branch
            _add_trigger(schema_editor, model, self.trigger)

    def database_backwards(
        self,
        app_label: str,
        schema_editor: BaseDatabaseSchemaEditor,
        from_state: ProjectState,
        to_state: ProjectState,
    ) -> None:
        model = to_state.apps.get_model(app_label, self.model_name)
        assert issubclass(model, Model)
        if self.allow_migrate_model_trigger(schema_editor, model):  # pragma: no branch
            _remove_trigger(schema_editor, model, self.trigger)

    def describe(self) -> str:
        return f"Create trigger {self.trigger.name} on model {self.model_name}"

    def deconstruct(self) -> tuple[str, Sequence[Any], dict[str, Any]]:
        return (
            self.__class__.__name__,
            [],
            {
                "model_name": self.model_name,
                "trigger": self.trigger,
            },
        )

    @property
    def migration_name_fragment(self) -> str:
        assert self.trigger.name is not None
        return f"{self.model_name_lower}_{self.trigger.name.lower()}"


def _get_trigger_by_name(model_state: ModelState, name: str) -> compiler.Trigger:
    for trigger in model_state.options.get("triggers", []):  # pragma: no branch
        assert isinstance(trigger, compiler.Trigger)
        if trigger.name == name:
            return trigger

    raise ValueError(f"No trigger named {name} on model {model_state.name}")  # pragma: no cover


class RemoveTrigger(TriggerOperationMixin, IndexOperation):
    option_name = "triggers"

    def __init__(self, model_name: str, name: str) -> None:
        self.model_name = model_name
        self.name = name

    def state_forwards(self, app_label: str, state: ProjectState) -> None:
        model_state = state.models[app_label, self.model_name]
        objs = model_state.options.get("triggers", [])
        model_state.options["triggers"] = [obj for obj in objs if obj.name != self.name]
        state.reload_model(app_label, self.model_name, delay=True)

    def database_forwards(
        self,
        app_label: str,
        schema_editor: BaseDatabaseSchemaEditor,
        from_state: ProjectState,
        to_state: ProjectState,
    ) -> None:
        model = to_state.apps.get_model(app_label, self.model_name)
        assert issubclass(model, Model)
        if self.allow_migrate_model_trigger(schema_editor, model):  # pragma: no branch
            from_model_state = from_state.models[app_label, self.model_name_lower]
            trigger = _get_trigger_by_name(from_model_state, self.name)
            _remove_trigger(schema_editor, model, trigger)

    def database_backwards(
        self,
        app_label: str,
        schema_editor: BaseDatabaseSchemaEditor,
        from_state: ProjectState,
        to_state: ProjectState,
    ) -> None:
        model = to_state.apps.get_model(app_label, self.model_name)
        assert issubclass(model, Model)
        if self.allow_migrate_model_trigger(schema_editor, model):  # pragma: no branch
            to_model_state = to_state.models[app_label, self.model_name_lower]
            trigger = _get_trigger_by_name(to_model_state, self.name)
            _add_trigger(schema_editor, model, trigger)

    def describe(self) -> str:
        return f"Remove trigger {self.name} from model {self.model_name}"

    def deconstruct(self) -> tuple[str, Sequence[Any], dict[str, Any]]:
        return (
            self.__class__.__name__,
            [],
            {
                "model_name": self.model_name,
                "name": self.name,
            },
        )

    @property
    def migration_name_fragment(self) -> str:
        return f"remove_{self.model_name_lower}_{self.name.lower()}"


def _inject_m2m_dependency_in_proxy(proxy_op):
    """
    Django does not properly add dependencies to m2m fields that are base classes for
    proxy models. Inject the dependency here
    """
    for base in proxy_op.bases:
        # Ignore inherited bases that are not models or abstract
        # https://github.com/Opus10/django-pgtrigger/issues/126
        if not (isinstance(base, str) and "." in base):
            continue

        model = apps.get_model(base)
        creator = model._meta.auto_created
        if creator:
            for field in creator._meta.many_to_many:
                if field.remote_field.through == model:
                    app_label, model_name = creator._meta.label_lower.split(".")
                    proxy_op._auto_deps.append(
                        OperationDependency(
                            app_label, model_name, field.name, OPERATION_DEPENDENCY_CREATE
                        )
                    )


class MigrationAutodetectorMixin:
    """A mixin that can be subclassed with MigrationAutodetector and detects triggers"""

    def _detect_changes(self, *args, **kwargs):
        self.altered_triggers = {}
        return super()._detect_changes(*args, **kwargs)

    def _get_add_trigger_op(self, model, trigger):
        if not isinstance(trigger, compiler.Trigger):
            trigger = trigger.compile(model)

        return AddTrigger(model_name=model._meta.model_name, trigger=trigger)

    def create_altered_constraints(self):
        """
        Piggyback off of constraint generation hooks to generate
        trigger migration operations
        """
        for app_label, model_name in sorted(self.kept_model_keys | self.kept_proxy_keys):
            old_model_name = self.renamed_models.get((app_label, model_name), model_name)
            old_model_state = self.from_state.models[app_label, old_model_name]
            new_model_state = self.to_state.models[app_label, model_name]
            new_model = self.to_state.apps.get_model(app_label, model_name)

            old_triggers = old_model_state.options.get("triggers", [])
            new_triggers = [
                trigger.compile(new_model)
                for trigger in new_model_state.options.get("triggers", [])
            ]
            add_triggers = [c for c in new_triggers if c not in old_triggers]
            rem_triggers = [c for c in old_triggers if c not in new_triggers]

            self.altered_triggers.update(
                {
                    (app_label, model_name): {
                        "added_triggers": add_triggers,
                        "removed_triggers": rem_triggers,
                    }
                }
            )

        return super().create_altered_constraints()

    def generate_added_constraints(self):
        for (app_label, model_name), alt_triggers in self.altered_triggers.items():
            model = self.to_state.apps.get_model(app_label, model_name)
            for trigger in alt_triggers["added_triggers"]:
                self.add_operation(
                    app_label, self._get_add_trigger_op(model=model, trigger=trigger)
                )

        return super().generate_added_constraints()

    def generate_removed_constraints(self):
        for (app_label, model_name), alt_triggers in self.altered_triggers.items():
            for trigger in alt_triggers["removed_triggers"]:
                self.add_operation(
                    app_label, RemoveTrigger(model_name=model_name, name=trigger.name)
                )

        return super().generate_removed_constraints()

    def generate_created_models(self):
        super().generate_created_models()

        added_models = self.new_model_keys - self.old_model_keys
        added_models = sorted(added_models, key=self.swappable_first_key, reverse=True)

        for app_label, model_name in added_models:
            model = self.to_state.apps.get_model(app_label, model_name)
            model_state = self.to_state.models[app_label, model_name]

            if not model_state.options.get("managed", True):
                continue  # pragma: no cover

            related_fields = {
                op.name: op.field
                for op in self.generated_operations.get(app_label, [])
                if isinstance(op, AddField) and model_name == op.model_name
            }

            related_dependencies = [
                OperationDependency(app_label, model_name, name, OPERATION_DEPENDENCY_CREATE)
                for name in sorted(related_fields)
            ]
            # Depend on the model being created
            related_dependencies.append(
                OperationDependency(app_label, model_name, None, OPERATION_DEPENDENCY_CREATE)
            )

            for trigger in model_state.options.pop("triggers", []):
                self.add_operation(
                    app_label,
                    self._get_add_trigger_op(model=model, trigger=trigger),
                    dependencies=related_dependencies,
                )

    def generate_created_proxies(self):
        super().generate_created_proxies()

        added = self.new_proxy_keys - self.old_proxy_keys
        for app_label, model_name in sorted(added):
            # Django has a bug that prevents it from injecting a dependency
            # to an M2M through model when a proxy model inherits it.
            # Inject an additional dependency for created proxies to
            # avoid this
            for op in self.generated_operations.get(app_label, []):
                if isinstance(op, CreateModel) and op.options.get(  # pragma: no branch
                    "proxy", False
                ):
                    _inject_m2m_dependency_in_proxy(op)

            model = self.to_state.apps.get_model(app_label, model_name)
            model_state = self.to_state.models[app_label, model_name]
            assert model_state.options.get("proxy")

            for trigger in model_state.options.pop("triggers", []):
                self.add_operation(
                    app_label,
                    self._get_add_trigger_op(model=model, trigger=trigger),
                    dependencies=[
                        OperationDependency(
                            app_label, model_name, None, OPERATION_DEPENDENCY_CREATE
                        )
                    ],
                )

    def generate_deleted_proxies(self):
        deleted = self.old_proxy_keys - self.new_proxy_keys
        for app_label, model_name in sorted(deleted):
            model_state = self.from_state.models[app_label, model_name]
            assert model_state.options.get("proxy")

            for trigger in model_state.options.pop("triggers", []):
                self.add_operation(
                    app_label,
                    RemoveTrigger(model_name=model_name, name=trigger.name),
                    dependencies=[
                        OperationDependency(
                            app_label, model_name, None, OPERATION_DEPENDENCY_CREATE
                        )
                    ],
                )

        super().generate_deleted_proxies()


class DatabaseSchemaEditorMixin:
    """
    A schema editor mixin that can subclass a DatabaseSchemaEditor and
    handle altering column types of triggers.

    Postgres does not allow altering column types of columns used in trigger
    conditions. Here we fix this with the following approach:

    1. Detect that a column type is being changed and set a flag so that we
       can alter behavior of the schema editor.
    2. In execute(), check for the special error message that's raised
       when trying to alter a column of a trigger. Temporarily drop triggers
       during the alter statement and reinstall them. Ensure this is all
       wrapped in a transaction
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.temporarily_dropped_triggers = set()
        self.is_altering_field_type = False

    @contextlib.contextmanager
    def alter_field_type(self):
        """
        Temporarily sets state so that execute() knows we are trying to alter a column type
        """
        self.is_altering_field_type = True
        try:
            yield
        finally:
            self.is_altering_field_type = False

    def _alter_field(
        self,
        model,
        old_field,
        new_field,
        old_type,
        new_type,
        old_db_params,
        new_db_params,
        strict=False,
    ):
        """
        Detects that a field type is being altered and sets the appropriate state
        """
        context = self.alter_field_type() if old_type != new_type else contextlib.nullcontext()

        with context:
            return super()._alter_field(
                model,
                old_field,
                new_field,
                old_type,
                new_type,
                old_db_params,
                new_db_params,
                strict=strict,
            )

    @contextlib.contextmanager
    def temporarily_drop_trigger(self, trigger, table):
        """
        Given a table and trigger, temporarily drop the trigger and recreate it
        after the context manager yields.
        """
        self.temporarily_dropped_triggers.add((trigger, table))

        try:
            with transaction.atomic(self.connection.alias), self.connection.cursor() as cursor:
                cursor.execute(
                    f"""
                    SELECT pg_get_triggerdef(oid) FROM pg_trigger
                        WHERE tgname = '{trigger}' AND tgrelid = '{table}'::regclass;
                """
                )
                trigger_create_sql = cursor.fetchall()[0][0]
                cursor.execute(f"DROP TRIGGER {trigger} on {utils.quote(table)};")

                yield

                cursor.execute(trigger_create_sql)
        finally:
            self.temporarily_dropped_triggers.remove((trigger, table))

    def execute(self, *args, **kwargs):
        """
        If we are altering a field type, catch the special error psycopg raises
        when a column on a trigger is altered. Temporarily drop and recreate
        triggers to ensure the alter operation is successful.
        """
        if self.is_altering_field_type:
            try:
                with transaction.atomic(self.connection.alias):
                    return super().execute(*args, **kwargs)
            except Exception as exc:
                match = re.search(
                    r"cannot alter type of a column used in a trigger definition\n"
                    r'DETAIL:\s+trigger (?P<trigger>\w+).+on table "?(?P<table>\w+)"?',
                    str(exc),
                )
                if match:
                    trigger = match.groupdict()["trigger"]
                    table = match.groupdict()["table"]

                    # In practice we should never receive the same error message for
                    # the same trigger/table, but check anyways to avoid infinite
                    # recursion
                    if (
                        trigger.startswith("pgtrigger_")
                        and (table, trigger) not in self.temporarily_dropped_triggers
                    ):
                        with self.temporarily_drop_trigger(trigger, table):
                            return self.execute(*args, **kwargs)

                raise  # pragma: no cover
        else:
            return super().execute(*args, **kwargs)

    def create_model(self, model):
        """
        Create the model

        `Meta.triggers` isn't populated on the forwards `CreateTable` migration
        (as Triggers are only added to the migration state via `AddTrigger`
        operations). `Meta.triggers` may be populated when:

        - The backwards operation of a `RemoveTable` operation where there was
          still triggers defined in the model state when the table was
          removed.
        - Creating the tables of an unmigrated app when `run_syncdb` is
          supplied to the migrate command (or when running tests).
        """
        super().create_model(model)

        for trigger in getattr(model._meta, "triggers", []):
            _add_trigger(self, model, trigger)



"""Tests multi-database support"""

import contextlib

import ddf
import pytest
from django.contrib.auth.models import User
from django.core.management import call_command
from django.db import transaction

import pgtrigger
from pgtrigger import core
from pgtrigger.tests import models, utils


class ToLogRouter:
    """
    Route the "ToLog" model to the "other" database
    """

    route_app_labels = {"auth", "contenttypes"}

    def db_for_write(self, model, **hints):
        if model == models.ToLogModel:
            return "other"

        return None


@pytest.fixture(autouse=True)
def routed_db(settings):
    settings.DATABASE_ROUTERS = [
        "pgtrigger.tests.test_multi_db.ToLogRouter",
        "pgtrigger.tests.models.Router",
    ]


@pytest.mark.django_db(databases=["default", "sqlite", "other"], transaction=True)
def test_multi_db_ignore():
    """Tests ignoring triggers across multiple databases"""
    trigger = pgtrigger.Protect(operation=pgtrigger.Delete, name="protect_deletes")

    with contextlib.ExitStack() as stack:
        stack.enter_context(trigger.register(models.ToLogModel))
        stack.enter_context(trigger.register(User))
        stack.enter_context(trigger.install(models.ToLogModel, database="other"))
        stack.enter_context(trigger.install(User))

        with utils.raises_trigger_error(match="Cannot delete", database="other"):
            log = ddf.G(models.ToLogModel)
            log.delete()

        with utils.raises_trigger_error(match="Cannot delete"):
            user = ddf.G(User)
            user.delete()

        with transaction.atomic():
            with pgtrigger.ignore("tests.ToLogModel:protect_deletes", "auth.User:protect_deletes"):
                log = models.ToLogModel.objects.create()
                log.delete()
                user = ddf.G(User)
                user.delete()

            with utils.raises_trigger_error(match="Cannot delete"):
                user = User.objects.create(username="hi")
                user.delete()

            with utils.raises_trigger_error(match="Cannot delete", database="other"):
                log = models.ToLogModel.objects.create()
                log.delete()


@pytest.mark.django_db(databases=["default", "sqlite", "other"])
def test_full_ls(capsys):
    call_command("pgtrigger", "ls")
    captured = capsys.readouterr()
    lines = [line for line in captured.out.split("\n") if line]
    for line in lines:
        assert "\x1b[92mINSTALLED\x1b[0m" in line

    call_command("pgtrigger", "ls", "-d", "other")
    captured = capsys.readouterr()
    lines = [line for line in captured.out.split("\n") if line]
    for line in lines:
        # The router ignores partition models for the default DB
        if "tests.PartitionModel:protect_delete" in line:
            assert "\x1b[94mUNALLOWED\x1b[0m" in line
        else:
            assert "\x1b[92mINSTALLED\x1b[0m" in line

    call_command("pgtrigger", "ls", "-d", "sqlite")
    captured = capsys.readouterr()
    lines = [line for line in captured.out.split("\n") if line]
    for line in lines:
        assert "\x1b[94mUNALLOWED\x1b[0m" in line


@pytest.mark.django_db(databases=["other"])
def test_disable_enable(capsys):
    call_command("pgtrigger", "disable", "-d", "other")
    for model, trigger in pgtrigger.registered():
        expected_status = None if model == models.PartitionModel else False
        assert trigger.get_installation_status(model, database="other")[1] is expected_status

    call_command("pgtrigger", "enable", "--database", "other")
    for model, trigger in pgtrigger.registered():
        expected_status = None if model == models.PartitionModel else True
        assert trigger.get_installation_status(model, database="other")[1] is expected_status


@pytest.mark.django_db(databases=["sqlite"])
def test_ignore_non_postgres_dbs():
    call_command("pgtrigger", "uninstall", "-d", "sqlite")
    call_command("pgtrigger", "install", "-d", "sqlite")
    call_command("pgtrigger", "install", "-d", "sqlite")
    call_command("pgtrigger", "prune", "-d", "sqlite")


@pytest.mark.django_db(databases=["other", "default", "sqlite"])
def test_uninstall_install():
    for model, trigger in pgtrigger.registered():
        expected_status = core.UNALLOWED if model == models.PartitionModel else core.INSTALLED
        assert trigger.get_installation_status(model, database="other")[0] == expected_status

    call_command("pgtrigger", "uninstall", "-d", "other")
    call_command("pgtrigger", "uninstall", "-d", "default")
    for model, trigger in pgtrigger.registered():
        expected_status = core.UNALLOWED if model == models.PartitionModel else core.UNINSTALLED
        assert trigger.get_installation_status(model, database="other")[0] == expected_status

    call_command("pgtrigger", "install", "--database", "other")
    for model, trigger in pgtrigger.registered():
        expecetd_status = core.UNALLOWED if model == models.PartitionModel else core.INSTALLED
        assert trigger.get_installation_status(model, database="other")[0] == expecetd_status

    for model, trigger in pgtrigger.registered():
        assert trigger.get_installation_status(model, database="default")[0] == core.UNINSTALLED



# Generated by Django 3.2.15 on 2022-08-17 22:11

from django.db import migrations
from psqlextra.backend.migrations.operations import PostgresAddRangePartition


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0010_auto_20220817_2211"),
    ]

    operations = [
        PostgresAddRangePartition(
            model_name="partitionmodel",
            name="pt1",
            from_values="2019-01-01",
            to_values="2019-02-01",
        ),
        PostgresAddRangePartition(
            model_name="partitionmodel",
            name="pt2",
            from_values="2019-02-01",
            to_values="2019-03-01",
        ),
        PostgresAddRangePartition(
            model_name="partitionmodel",
            name="pt3",
            from_values="2019-03-01",
            to_values="2019-04-01",
        ),
    ]



import django.apps
from django.db import connections
from django.db.models.signals import pre_migrate


def install_schemas(using, **kwargs):
    if connections[using].vendor == "postgresql":
        with connections[using].cursor() as cursor:
            cursor.execute('CREATE SCHEMA IF NOT EXISTS "order";')
            cursor.execute("CREATE SCHEMA IF NOT EXISTS receipt;")


class PGTriggerTestsConfig(django.apps.AppConfig):
    name = "pgtrigger.tests"

    def ready(self):
        """
        Ensure schemas are created for test databases before migrations
        """
        pre_migrate.connect(install_schemas, sender=self)



"""
The primary functional API for pgtrigger
"""

import logging
from typing import List, Tuple, Union

from django.db import DEFAULT_DB_ALIAS, connections

from pgtrigger import features, registry, utils

# The core pgtrigger logger
LOGGER = logging.getLogger("pgtrigger")


def install(*uris: str, database: Union[str, None] = None) -> None:
    """
    Install triggers.

    Args:
        *uris: URIs of triggers to install. If none are provided,
            all triggers are installed and orphaned triggers are pruned.
        database: The database. Defaults to the "default" database.
    """
    for model, trigger in registry.registered(*uris):
        LOGGER.info(
            "pgtrigger: Installing %s trigger for %s table on %s database.",
            trigger,
            model._meta.db_table,
            database or DEFAULT_DB_ALIAS,
        )
        trigger.install(model, database=database)

    if not uris and features.prune_on_install():  # pragma: no branch
        prune(database=database)


def prunable(database: Union[str, None] = None) -> List[Tuple[str, str, bool, str]]:
    """Return triggers that are candidates for pruning

    Args:
        database: The database. Defaults to the "default" database.

    Returns:
        A list of tuples consisting of the table, trigger ID, enablement, and database
    """
    if not utils.is_postgres(database):
        return []

    registered = {
        (utils.quote(model._meta.db_table), trigger.get_pgid(model))
        for model, trigger in registry.registered()
    }

    with utils.connection(database).cursor() as cursor:
        parent_trigger_clause = "tgparentid = 0 AND" if utils.pg_maj_version(cursor) >= 13 else ""

        # Only select triggers that are in the current search path. We accomplish
        # this by parsing the tgrelid and only selecting triggers that don't have
        # a schema name in their path
        cursor.execute(
            f"""
            SELECT tgrelid::regclass, tgname, tgenabled
                FROM pg_trigger
                WHERE tgname LIKE 'pgtrigger_%%' AND
                      {parent_trigger_clause}
                      array_length(parse_ident(tgrelid::regclass::varchar), 1) = 1
            """
        )
        triggers = set(cursor.fetchall())

    return [
        (trigger[0], trigger[1], trigger[2] == "O", database or DEFAULT_DB_ALIAS)
        for trigger in triggers
        if (utils.quote(trigger[0]), trigger[1]) not in registered
    ]


def prune(database: Union[str, None] = None) -> None:
    """
    Remove any pgtrigger triggers in the database that are not used by models.
    I.e. if a model or trigger definition is deleted from a model, ensure
    it is removed from the database

    Args:
        database: The database. Defaults to the "default" database.
    """
    for trigger in prunable(database=database):
        LOGGER.info(
            "pgtrigger: Pruning trigger %s for table %s on %s database.",
            trigger[1],
            trigger[0],
            trigger[3],
        )

        connection = connections[trigger[3]]
        uninstall_sql = utils.render_uninstall(trigger[0], trigger[1])
        with connection.cursor() as cursor:
            cursor.execute(uninstall_sql)


def enable(*uris: str, database: Union[str, None] = None) -> None:
    """
    Enables registered triggers.

    Args:
        *uris: URIs of triggers to enable. If none are provided,
            all triggers are enabled.
        database: The database. Defaults to the "default" database.
    """
    for model, trigger in registry.registered(*uris):
        LOGGER.info(
            "pgtrigger: Enabling %s trigger for %s table on %s database.",
            trigger,
            model._meta.db_table,
            database or DEFAULT_DB_ALIAS,
        )
        trigger.enable(model, database=database)


def uninstall(*uris: str, database: Union[str, None] = None) -> None:
    """
    Uninstalls triggers.

    Args:
        *uris: URIs of triggers to uninstall. If none are provided,
            all triggers are uninstalled and orphaned triggers are pruned.
        database: The database. Defaults to the "default" database.
    """
    for model, trigger in registry.registered(*uris):
        LOGGER.info(
            "pgtrigger: Uninstalling %s trigger for %s table on %s database.",
            trigger,
            model._meta.db_table,
            database or DEFAULT_DB_ALIAS,
        )
        trigger.uninstall(model, database=database)

    if not uris and features.prune_on_install():
        prune(database=database)


def disable(*uris: str, database: Union[str, None] = None) -> None:
    """
    Disables triggers.

    Args:
        *uris: URIs of triggers to disable. If none are provided,
            all triggers are disabled.
        database: The database. Defaults to the "default" database.
    """
    for model, trigger in registry.registered(*uris):
        LOGGER.info(
            "pgtrigger: Disabling %s trigger for %s table on %s database.",
            trigger,
            model._meta.db_table,
            database or DEFAULT_DB_ALIAS,
        )
        trigger.disable(model, database=database)



import datetime as dt

import ddf
import pytest
from django.contrib.auth.models import User
from django.db import transaction
from django.db.utils import NotSupportedError

import pgtrigger
from pgtrigger import core
from pgtrigger.tests import models, utils


def test_func():
    """Tests using custom Func object"""
    trigger = pgtrigger.Trigger(
        name="example",
        when=pgtrigger.After,
        operation=pgtrigger.Delete,
        func=pgtrigger.Func("SELECT {columns.int_field} FROM {meta.db_table}"),
    )
    assert trigger.render_func(models.TestModel) == "SELECT int_field FROM tests_testmodel"


@pytest.mark.django_db
def test_partition():
    p1 = ddf.G(models.PartitionModel, timestamp=dt.datetime(2019, 1, 3))
    p2 = ddf.G(models.PartitionModel, timestamp=dt.datetime(2019, 2, 4))
    p3 = ddf.G(models.PartitionModel, timestamp=dt.datetime(2019, 3, 5))
    default = ddf.G(models.PartitionModel, timestamp=dt.datetime(2019, 4, 5))

    with utils.raises_trigger_error(match="Cannot delete"):
        p1.delete()

    with utils.raises_trigger_error(match="Cannot delete"):
        p2.delete()

    with utils.raises_trigger_error(match="Cannot delete"):
        p3.delete()

    with utils.raises_trigger_error(match="Cannot delete"):
        default.delete()

    with pgtrigger.ignore("tests.PartitionModel:protect_delete"):
        p1.delete()
        p2.delete()
        p3.delete()
        default.delete()


@pytest.mark.django_db
def test_through_model():
    """
    Tests the "ThroughTrigger" model to verify that triggers execute on M2M through models
    """
    test_trigger = ddf.G(models.TestTrigger)
    test_trigger.m2m_field.add(ddf.G("auth.User"))

    with utils.raises_trigger_error(match="Cannot delete"):
        test_trigger.m2m_field.clear()


@pytest.mark.django_db
def test_statement_row_level_logging():
    """
    Updates "ToLogModel" entries, which have statement, row-level,
    and referencing statement triggers that create log entries.
    """
    ddf.G(models.ToLogModel, n=5, field="old_field")

    assert not models.LogEntry.objects.exists()

    models.ToLogModel.objects.update(field="new_field")

    # The statement-level trigger without references should have produced
    # one log entry
    assert models.LogEntry.objects.filter(level="STATEMENT", old_field__isnull=True).count() == 1

    # The statement-level trigger with references should have made log
    # entries for all of the old values and the new updated values
    assert models.LogEntry.objects.filter(level="STATEMENT", old_field__isnull=False).count() == 5
    assert (
        models.LogEntry.objects.filter(
            level="STATEMENT", old_field="old_field", new_field="new_field"
        ).count()
        == 5
    )

    # The row-level trigger should have produced five entries
    assert models.LogEntry.objects.filter(level="ROW").count() == 5
    obj = models.ToLogModel.objects.first()
    obj.save()

    # A duplicate update shouldn't fire any more row-level log entries
    assert models.LogEntry.objects.filter(level="ROW").count() == 5


@pytest.mark.django_db(transaction=True)
def test_deferred_trigger():
    """
    Tests deferrable execution of a trigger
    """
    # Make the LogEntry model a soft delete model where
    # "level" is set to "inactive"
    trigger = pgtrigger.Protect(
        name="protect_delete",
        when=pgtrigger.After,
        operation=pgtrigger.Delete,
        timing=pgtrigger.Deferred,
    )
    with trigger.register(models.TestModel), trigger.install(models.TestModel):
        obj = ddf.G(models.TestModel)
        deferring_worked = False
        with utils.raises_trigger_error(match="Cannot delete"):
            with transaction.atomic():
                obj.delete()
                # Deletion works within the transaction, but fails
                # when the transaction commits.
                assert not models.TestModel.objects.exists()
                deferring_worked = True

        assert deferring_worked
        assert models.TestModel.objects.exists()
        obj = models.TestModel.objects.get()

        # Verify that we can ignore deferrable triggers
        with pgtrigger.ignore("tests.TestModel:protect_delete"):
            with transaction.atomic():
                obj.delete()
                assert not models.TestModel.objects.exists()

            # The object should still not exist outside of the transaction
            assert not models.TestModel.objects.exists()


@pytest.mark.django_db
def test_updating_trigger_condition():
    """
    Tests re-installing a trigger when the condition changes
    """
    # Make the LogEntry model a soft delete model where
    # "level" is set to "inactive"
    trigger = pgtrigger.Protect(name="protect_delete", operation=pgtrigger.Delete)
    with trigger.install(models.LogEntry):
        le = ddf.G(models.LogEntry, level="good")

        with utils.raises_trigger_error(match="Cannot delete"):
            le.delete()

        # Protect deletes when "level" is "bad". The trigger should be reinstalled
        # appropriately
        trigger.condition = pgtrigger.Q(old__level="bad")
        with trigger.install(models.LogEntry):
            le.delete()


def test_declaration_rendering():
    """Verifies that triggers with a DECLARE are rendered correctly"""

    class DeclaredTrigger(pgtrigger.Trigger):
        def get_declare(self, model):
            return [("var_name", "UUID")]

    rendered = DeclaredTrigger(
        name="test", when=pgtrigger.Before, operation=pgtrigger.Insert
    ).render_declare(None)
    assert rendered == "DECLARE var_name UUID;"

    class DeclaredTriggerMultiple(pgtrigger.Trigger):
        def get_declare(self, model):
            return [("var_name", "UUID"), ("var2_name", "UUID")]

    rendered = DeclaredTriggerMultiple(
        name="test", when=pgtrigger.Before, operation=pgtrigger.Insert
    ).render_declare(None)
    assert rendered == "DECLARE var_name UUID; var2_name UUID;"


def test_f():
    """Tests various properties of the pgtrigger.F object"""
    with pytest.raises(ValueError, match="must reference"):
        pgtrigger.F("bad_value")

    assert pgtrigger.F("old__value").resolved_name == 'OLD."value"'


@pytest.mark.django_db
def test_is_distinct_from_condition():
    """Tests triggers where the old and new are distinct from one another

    Note that distinct is the not the same as not being equal since nulls
    are never equal
    """
    test_model = ddf.G(models.TestTrigger, int_field=0)

    # Protect a field from being updated to a different value
    trigger = pgtrigger.Protect(
        name="protect",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(old__int_field__df=pgtrigger.F("new__int_field"))
        | pgtrigger.Q(new__nullable__df=pgtrigger.F("old__nullable")),
    )
    with trigger.install(models.TestTrigger):
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_model.int_field = 1
            test_model.save()

        # Ensure the null case works
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_model.nullable = "1"
            test_model.save()

        # Saving the same values should work fine
        test_model.int_field = 0
        test_model.nullable = None
        test_model.save()


@pytest.mark.django_db
def test_invalid_trigger():
    """Ensures triggers with invalid syntax are not installed"""
    # Truncates can only be used on statement level triggers
    trigger = pgtrigger.Protect(
        name="test_invalid",
        operation=pgtrigger.Truncate,
    )
    with pytest.raises(NotSupportedError, match="are not supported"):
        trigger.install(models.TestTrigger)


@pytest.mark.django_db
def test_is_distinct_from_condition_fk_field():
    """Tests triggers where the old and new are distinct from one another
    on a foreign key field

    Django doesnt support custom lookups by default, and this tests some
    of the overridden behavior
    """
    test_int_fk_model = ddf.G(models.TestTrigger, fk_field=None)

    # Protect a foreign key from being updated to a different value
    trigger = pgtrigger.Protect(
        name="test_is_distinct_from_condition_fk_field1",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(old__fk_field__df=pgtrigger.F("new__fk_field")),
    )
    with trigger.install(models.TestTrigger):
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_int_fk_model.fk_field = User(id=1)
            test_int_fk_model.save()

        # Saving the same values should work fine
        test_int_fk_model.fk_field = None
        test_int_fk_model.save()

    # Protect a non-int foreign key from being updated to a different value
    char_pk = ddf.G(models.CharPk)
    test_char_fk_model = ddf.G(models.TestTrigger, char_pk_fk_field=char_pk)
    trigger = pgtrigger.Protect(
        name="test_is_distinct_from_condition_fk_field2",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(old__char_pk_fk_field__df=pgtrigger.F("new__char_pk_fk_field")),
    )
    with trigger.install(models.TestTrigger):
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_char_fk_model.char_pk_fk_field = None
            test_char_fk_model.save()

        # Saving the same values should work fine
        test_char_fk_model.char_pk_fk_field = char_pk
        test_char_fk_model.save()


@pytest.mark.django_db
def test_is_not_distinct_from_condition():
    """Tests triggers where the old and new are not distinct from one another

    Note that distinct is the not the same as not being equal since nulls
    are never equal
    """
    test_model = ddf.G(models.TestTrigger, int_field=0)

    # Protect a field from being updated to the same value. In this case,
    # both int_field and nullable need to change in order for the update to
    # happen
    trigger = pgtrigger.Protect(
        name="test_is_not_distinct_from_condition1",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(old__int_field__ndf=pgtrigger.F("new__int_field"))
        | pgtrigger.Q(old__nullable__ndf=pgtrigger.F("new__nullable")),
    )
    with trigger.install(models.TestTrigger):
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_model.int_field = 1
            test_model.save()

        # Ensure the null case works
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_model.int_field = 0
            test_model.nullable = "1"
            test_model.save()

        # Updating both fields will ignore the trigger
        test_model.int_field = 1
        test_model.nullable = "1"
        test_model.save()


def test_max_name_length(mocker):
    """
    Verifies that a trigger with the exact MAX_NAME_LENGTH can be installed
    fine. Also checks that going above this by one character results in
    a database error
    """
    # Protect a field from being updated to the same value. In this case,
    # both int_field and nullable need to change in order for the update to
    # happen
    trigger = pgtrigger.Protect(
        name="t" * core.MAX_NAME_LENGTH,
        operation=pgtrigger.Update,
    )
    assert trigger.get_pgid(models.TestTrigger)

    mocker.patch.object(pgtrigger.Protect, "validate_name")
    with pytest.raises(ValueError):
        trigger = pgtrigger.Protect(
            name="a" * (core.MAX_NAME_LENGTH + 1),
            operation=pgtrigger.Update,
        )
        trigger.get_pgid(models.TestTrigger)


def test_invalid_name_characters(mocker):
    """
    Verifies that trigger names must contain only alphanumeric
    characters, hyphens, or underscores
    """
    pgtrigger.Protect(
        name="hello_world-111",
        operation=pgtrigger.Update,
    )
    with pytest.raises(ValueError, match="alphanumeric"):
        pgtrigger.Protect(
            name="hello.world",
            operation=pgtrigger.Update,
        )


@pytest.mark.django_db
def test_complex_conditions():
    """Tests complex OLD and NEW trigger conditions"""
    zero_to_one = ddf.G(models.TestModel, int_field=0)

    # Dont let intfield go from 0 -> 1
    trigger = pgtrigger.Protect(
        name="test_complex_conditions1",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(old__int_field=0, new__int_field=1),
    )
    with trigger.install(models.TestModel):
        with utils.raises_trigger_error(match="Cannot update rows"):
            zero_to_one.int_field = 1
            zero_to_one.save()

    # Test a condition with a datetime field
    test_model = ddf.G(models.TestTrigger, int_field=0, dt_field=dt.datetime(2020, 1, 1))
    trigger = pgtrigger.Protect(
        name="test_complex_conditions2",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        condition=(
            pgtrigger.Q(old__int_field=0, new__int_field=1)
            | pgtrigger.Q(new__dt_field__lt=dt.datetime(2020, 1, 1))
        ),
    )
    with trigger.install(models.TestTrigger):
        with utils.raises_trigger_error(match="Cannot update rows"):
            test_model.int_field = 1
            test_model.save()
        test_model.int_field = 2
        test_model.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            test_model.dt_field = dt.datetime(2019, 1, 1)
            test_model.save()


def test_referencing_rendering():
    """Verifies the rendering of the Referencing construct"""
    assert (
        str(pgtrigger.Referencing(old="old_table")).strip() == "REFERENCING OLD TABLE AS old_table"
    )
    assert (
        str(pgtrigger.Referencing(new="new_table")).strip() == "REFERENCING NEW TABLE AS new_table"
    )
    assert (
        str(pgtrigger.Referencing(old="old_table", new="new_table")).strip()
        == "REFERENCING OLD TABLE AS old_table  NEW TABLE AS new_table"
    )


def test_arg_checks():
    """
    There are quite a few places that check arguments in the trigger module.
    Enumerate these cases here to make sure they work
    """

    with pytest.raises(ValueError, match='Must provide either "old" and/or "new"'):
        pgtrigger.Referencing()

    with pytest.raises(ValueError, match="Must provide SQL"):
        pgtrigger.Condition()

    with pytest.raises(ValueError, match="Must provide at least one"):
        pgtrigger.UpdateOf()

    with pytest.raises(ValueError, match='must have "name"'):
        pgtrigger.Trigger(when=pgtrigger.Before, operation=pgtrigger.Update)

    with pytest.raises(ValueError, match='Invalid "level"'):
        pgtrigger.Trigger(level="invalid")

    with pytest.raises(ValueError, match='Invalid "when"'):
        pgtrigger.Trigger(when="invalid")

    with pytest.raises(ValueError, match='Invalid "operation"'):
        pgtrigger.Trigger(when=pgtrigger.Before, operation="invalid")

    with pytest.raises(ValueError, match='Invalid "timing"'):
        pgtrigger.Trigger(when=pgtrigger.Before, operation=pgtrigger.Update, timing="timing")

    with pytest.raises(ValueError, match="Row-level triggers cannot have"):
        pgtrigger.Trigger(
            when=pgtrigger.Before,
            operation=pgtrigger.Update,
            referencing=pgtrigger.Referencing(old="old_table"),
        )

    with pytest.raises(ValueError, match='must have "level" attribute'):
        pgtrigger.Trigger(
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            timing=pgtrigger.Immediate,
        )

    with pytest.raises(ValueError, match='must have "when" attribute'):
        pgtrigger.Trigger(
            level=pgtrigger.Row,
            when=pgtrigger.Before,
            operation=pgtrigger.Update,
            timing=pgtrigger.Immediate,
        )

    with pytest.raises(ValueError, match="Must define func"):
        pgtrigger.Trigger(name="test", when=pgtrigger.Before, operation=pgtrigger.Update).get_func(
            None
        )

    with pytest.raises(ValueError, match="> 47"):
        pgtrigger.Trigger(  # noqa
            when=pgtrigger.Before, operation=pgtrigger.Update, name="1" * 48
        ).pgid


def test_operations():
    """Tests Operation objects and ORing them together"""
    assert str(pgtrigger.Update) == "UPDATE"
    assert str(pgtrigger.UpdateOf("col1")) == 'UPDATE OF "col1"'
    assert str(pgtrigger.UpdateOf("c1", "c2")) == 'UPDATE OF "c1", "c2"'

    assert str(pgtrigger.Update | pgtrigger.Delete) == "UPDATE OR DELETE"
    assert (
        str(pgtrigger.Update | pgtrigger.Delete | pgtrigger.Insert) == "UPDATE OR DELETE OR INSERT"
    )
    assert str(pgtrigger.Delete | pgtrigger.Update) == "DELETE OR UPDATE"


@pytest.mark.django_db
def test_custom_trigger_definitions():
    """Test a variety of custom trigger definitions"""
    test_model = ddf.G(models.TestTrigger)

    # Protect against inserts or updates
    # Note: Although we could use the "protect" trigger for this,
    # we manually provide the trigger code to test manual declarations
    trigger = pgtrigger.Trigger(
        name="test_custom_definition1",
        when=pgtrigger.Before,
        operation=pgtrigger.Insert | pgtrigger.Update,
        func="RAISE EXCEPTION 'no no no!';",
    )
    with trigger.install(test_model):
        # Inserts and updates are no longer available
        with utils.raises_trigger_error(match="no no no!"):
            models.TestTrigger.objects.create()

        with utils.raises_trigger_error(match="no no no!"):
            test_model.save()

    # Inserts and updates should work again
    ddf.G(models.TestTrigger)
    test_model.save()

    # Protect updates of a single column
    trigger = pgtrigger.Trigger(
        name="test_custom_definition2",
        when=pgtrigger.Before,
        operation=pgtrigger.UpdateOf("int_field"),
        func="RAISE EXCEPTION 'no no no!';",
    )
    with trigger.install(models.TestTrigger):
        # "field" should be able to be updated, but other_field should not
        test_model.save(update_fields=["field"])

        with utils.raises_trigger_error(match="no no no!"):
            test_model.save(update_fields=["int_field"])

    # Protect statement-level creates
    trigger = pgtrigger.Trigger(
        name="test_custom_definition3",
        level=pgtrigger.Statement,
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        func="RAISE EXCEPTION 'bad statement!';",
    )
    with trigger.install(models.TestTrigger):
        with utils.raises_trigger_error(match="bad statement!"):
            test_model.save()


@pytest.mark.django_db
def test_trigger_conditions():
    """Tests triggers with custom conditions"""
    test_model = ddf.G(models.TestTrigger)

    # Protect against inserts only when "field" is "hello"
    trigger = pgtrigger.Trigger(
        name="test_condition1",
        when=pgtrigger.Before,
        operation=pgtrigger.Insert,
        func="RAISE EXCEPTION 'no no no!';",
        condition=pgtrigger.Q(new__field="hello"),
    )
    with trigger.install(test_model):
        ddf.G(models.TestTrigger, field="hi!")
        with utils.raises_trigger_error(match="no no no!"):
            models.TestTrigger.objects.create(field="hello")

    # Protect updates where nothing is actually updated
    trigger = pgtrigger.Trigger(
        name="test_condition2",
        when=pgtrigger.Before,
        operation=pgtrigger.Update,
        func="RAISE EXCEPTION 'no no no!';",
        condition=pgtrigger.Condition("OLD.* IS NOT DISTINCT FROM NEW.*"),
    )
    with trigger.install(test_model):
        test_model.int_field = test_model.int_field + 1
        test_model.save()

        # Saving the same fields again will cause an error
        with utils.raises_trigger_error(match="no no no!"):
            test_model.save()

    # Make a model readonly when the int_field is -1
    read_only = ddf.G(models.TestModel, int_field=-1)
    non_read_only = ddf.G(models.TestModel, int_field=-2)

    trigger = pgtrigger.Trigger(
        name="test_condition3",
        when=pgtrigger.Before,
        operation=pgtrigger.Update | pgtrigger.Delete,
        func="RAISE EXCEPTION 'no no no!';",
        condition=pgtrigger.Q(old__int_field=-1),
    )
    with trigger.install(models.TestModel):
        with utils.raises_trigger_error(match="no no no!"):
            read_only.save()

        with utils.raises_trigger_error(match="no no no!"):
            read_only.delete()

        non_read_only.save()
        non_read_only.delete()


@pytest.mark.parametrize(
    "condition, expected_sql",
    [
        (pgtrigger.AnyChange(), "OLD.* IS DISTINCT FROM NEW.*"),
        (pgtrigger.AnyDontChange(), "OLD.* IS NOT DISTINCT FROM NEW.*"),
        (
            pgtrigger.AllChange(),
            '(OLD."char_pk_fk_field_id" IS DISTINCT FROM (NEW."char_pk_fk_field_id") AND OLD."dt_field" IS DISTINCT FROM (NEW."dt_field") AND OLD."field" IS DISTINCT FROM (NEW."field") AND OLD."fk_field_id" IS DISTINCT FROM (NEW."fk_field_id") AND OLD."id" IS DISTINCT FROM (NEW."id") AND OLD."int_field" IS DISTINCT FROM (NEW."int_field") AND OLD."nullable" IS DISTINCT FROM (NEW."nullable"))',  # noqa
        ),
        (
            pgtrigger.AllDontChange(),
            '(OLD."char_pk_fk_field_id" IS NOT DISTINCT FROM (NEW."char_pk_fk_field_id") AND OLD."dt_field" IS NOT DISTINCT FROM (NEW."dt_field") AND OLD."field" IS NOT DISTINCT FROM (NEW."field") AND OLD."fk_field_id" IS NOT DISTINCT FROM (NEW."fk_field_id") AND OLD."id" IS NOT DISTINCT FROM (NEW."id") AND OLD."int_field" IS NOT DISTINCT FROM (NEW."int_field") AND OLD."nullable" IS NOT DISTINCT FROM (NEW."nullable"))',  # noqa
        ),
        (~pgtrigger.AnyChange(), "NOT (OLD.* IS DISTINCT FROM NEW.*)"),
        (
            ~pgtrigger.AllChange(),
            'NOT ((OLD."char_pk_fk_field_id" IS DISTINCT FROM (NEW."char_pk_fk_field_id") AND OLD."dt_field" IS DISTINCT FROM (NEW."dt_field") AND OLD."field" IS DISTINCT FROM (NEW."field") AND OLD."fk_field_id" IS DISTINCT FROM (NEW."fk_field_id") AND OLD."id" IS DISTINCT FROM (NEW."id") AND OLD."int_field" IS DISTINCT FROM (NEW."int_field") AND OLD."nullable" IS DISTINCT FROM (NEW."nullable")))',  # noqa
        ),
        (pgtrigger.AnyChange("field"), 'OLD."field" IS DISTINCT FROM (NEW."field")'),
        (pgtrigger.AllChange("field"), 'OLD."field" IS DISTINCT FROM (NEW."field")'),
        (~pgtrigger.AnyChange("field"), 'NOT (OLD."field" IS DISTINCT FROM (NEW."field"))'),
        (~pgtrigger.AllChange("field"), 'NOT (OLD."field" IS DISTINCT FROM (NEW."field"))'),
        (
            pgtrigger.AnyChange("int_field", "dt_field"),
            '(OLD."dt_field" IS DISTINCT FROM (NEW."dt_field") OR OLD."int_field" IS DISTINCT FROM (NEW."int_field"))',  # noqa
        ),
        (
            pgtrigger.AnyDontChange("dt_field", "fk_field"),
            '(OLD."dt_field" IS NOT DISTINCT FROM (NEW."dt_field") OR OLD."fk_field_id" IS NOT DISTINCT FROM (NEW."fk_field_id"))',  # noqa
        ),
        (
            pgtrigger.AllChange("int_field", "dt_field"),
            '(OLD."dt_field" IS DISTINCT FROM (NEW."dt_field") AND OLD."int_field" IS DISTINCT FROM (NEW."int_field"))',  # noqa
        ),
        (
            pgtrigger.AllDontChange("int_field", "dt_field"),
            '(OLD."dt_field" IS NOT DISTINCT FROM (NEW."dt_field") AND OLD."int_field" IS NOT DISTINCT FROM (NEW."int_field"))',  # noqa
        ),
        (
            pgtrigger.AnyChange("int_field", "dt_field", exclude=["int_field"]),
            'OLD."dt_field" IS DISTINCT FROM (NEW."dt_field")',
        ),
        (
            pgtrigger.AnyChange(
                "fk_field",
                "char_pk_fk_field",
                "int_field",
                "dt_field",
                exclude=["char_pk_fk_field", "field"],
                exclude_auto=True,
            ),
            '(OLD."fk_field_id" IS DISTINCT FROM (NEW."fk_field_id") OR OLD."int_field" IS DISTINCT FROM (NEW."int_field"))',  # noqa
        ),
        (
            pgtrigger.AnyChange(exclude_auto=True),
            '(OLD."char_pk_fk_field_id" IS DISTINCT FROM (NEW."char_pk_fk_field_id") OR OLD."field" IS DISTINCT FROM (NEW."field") OR OLD."fk_field_id" IS DISTINCT FROM (NEW."fk_field_id") OR OLD."id" IS DISTINCT FROM (NEW."id") OR OLD."int_field" IS DISTINCT FROM (NEW."int_field") OR OLD."nullable" IS DISTINCT FROM (NEW."nullable"))',  # noqa
        ),
        (
            pgtrigger.AllChange(exclude_auto=True),
            '(OLD."char_pk_fk_field_id" IS DISTINCT FROM (NEW."char_pk_fk_field_id") AND OLD."field" IS DISTINCT FROM (NEW."field") AND OLD."fk_field_id" IS DISTINCT FROM (NEW."fk_field_id") AND OLD."id" IS DISTINCT FROM (NEW."id") AND OLD."int_field" IS DISTINCT FROM (NEW."int_field") AND OLD."nullable" IS DISTINCT FROM (NEW."nullable"))',  # noqa
        ),
    ],
)
def test_changed_condition(condition, expected_sql):
    """Tests the pgtrigger.Changed condition utility"""
    assert condition.resolve(models.ChangedCondition) == expected_sql


def test_changed_condition_bad_field():
    """Verifies incorrect fields aren't allowed in changed conditions"""
    with pytest.raises(ValueError, match="not found on model"):
        pgtrigger.AnyChange("bad_field").resolve(models.ChangedCondition)

    with pytest.raises(ValueError, match="not found on model"):
        pgtrigger.AnyChange("m2m_field").resolve(models.ChangedCondition)

    with pytest.raises(ValueError, match="not found on model"):
        pgtrigger.AnyChange(exclude=["bad_field"]).resolve(models.ChangedCondition)


@pytest.mark.django_db(databases=["default", "other"], transaction=True)
@pytest.mark.order(-2)  # This is a leaky test that modifies the schema. Always run last
def test_trigger_management(mocker):
    """Verifies dropping and recreating triggers works"""
    deletion_protected_model = ddf.G(models.TestTrigger)

    # Triggers should be installed initially
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    # Deactivate triggers. Deletions should happen without issue.
    # Note: run twice for idempotency checks
    pgtrigger.disable()
    pgtrigger.disable()
    deletion_protected_model.delete()

    # Reactivate triggers. Deletions should be protected
    pgtrigger.enable()
    pgtrigger.enable()
    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    # Do the same tests again, except this time uninstall and reinstall
    # triggers
    pgtrigger.uninstall()
    pgtrigger.uninstall()
    deletion_protected_model.delete()

    # Reactivate triggers. Deletions should be protected
    pgtrigger.install()
    pgtrigger.install()
    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    # Pruning triggers should do nothing at the moment
    pgtrigger.prune()
    pgtrigger.prune()
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    # However, changing the trigger name will cause the old triggers to
    # be pruned
    mocker.patch(
        "pgtrigger.Protect.name",
        new_callable=mocker.PropertyMock,
        return_value="hi",
    )
    pgtrigger.prune()
    pgtrigger.prune()
    deletion_protected_model.delete()



from django.conf import settings


def model_meta():
    """
    True if model meta support is enabled
    """
    return getattr(settings, "PGTRIGGER_MODEL_META", True)


def schema_editor():
    """
    True if we are using the patched Postgres schema editor.

    Note that setting this to False means that we cannot easily
    alter columns of models that are associated with trigger
    conditions
    """
    return getattr(settings, "PGTRIGGER_SCHEMA_EDITOR", True)


def migrations():
    """
    True if migrations are enabled
    """
    return model_meta() and getattr(settings, "PGTRIGGER_MIGRATIONS", True)


def install_on_migrate():
    """
    True if triggers should be installed after migrations
    """
    return getattr(settings, "PGTRIGGER_INSTALL_ON_MIGRATE", False)


def schema():
    """
    The default schema where special objects are installed
    """
    return getattr(settings, "PGTRIGGER_SCHEMA", "public")


def prune_on_install():
    """
    True if triggers should be pruned on a full install or uninstall
    """
    return getattr(settings, "PGTRIGGER_PRUNE_ON_INSTALL", True)



# Generated by Django 3.0.7 on 2020-07-21 19:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0003_auto_20200718_0938"),
    ]

    operations = [
        migrations.CreateModel(
            name="FSM",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("transition", models.CharField(max_length=32)),
            ],
        ),
    ]



"""Tests multi-database support"""

# flake8: noqa

import contextlib

import ddf
from django.core.management import call_command
import pytest

import pgtrigger
from pgtrigger.tests import models, utils


class SchemaRouter:
    """
    A router to control tables that should be migrated to different schemas
    """

    def db_for_read(self, model, **hints):
        if model == models.OrderSchema:
            return "order"
        elif model == models.ReceiptSchema:  # pragma: no branch
            return "receipt"

    def db_for_write(self, model, **hints):
        return self.db_for_read(model, **hints)

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if model_name == "orderschema":
            return db == "order"
        elif model_name == "receiptschema":
            return db == "receipt"


@pytest.fixture(autouse=True)
def routed_db(settings):
    settings.DATABASE_ROUTERS = [
        "pgtrigger.tests.test_multi_schema.SchemaRouter",
        "pgtrigger.tests.models.Router",
    ]


@pytest.fixture(autouse=True)
def schema_triggers():
    protect_deletes = pgtrigger.Protect(name="protect_deletes", operation=pgtrigger.Delete)
    protect_updates = pgtrigger.Protect(name="protect_updates", operation=pgtrigger.Update)

    with contextlib.ExitStack() as stack:
        stack.enter_context(protect_deletes.register(models.OrderSchema))
        stack.enter_context(protect_updates.register(models.ReceiptSchema))

        yield


@pytest.mark.django_db(databases=["order", "receipt"], transaction=True)
def test_multi_schema_triggers_work():
    """Verify the triggers in the schema_triggers fixture work"""
    call_command("pgtrigger", "install", "-d", "order")
    call_command("pgtrigger", "install", "-d", "receipt")

    order = ddf.G("tests.OrderSchema")
    receipt = ddf.G("tests.ReceiptSchema")

    with utils.raises_trigger_error(match="Cannot delete", database="order"):
        order.delete()

    with utils.raises_trigger_error(match="Cannot update", database="receipt"):
        receipt.char_field = "hello"
        receipt.save()

    receipt.delete()

    order = ddf.G("tests.OrderSchema")
    with pgtrigger.ignore("tests.OrderSchema:protect_deletes"):
        order.delete()


@pytest.mark.django_db(databases=["order", "receipt", "default", "other"], transaction=True)
def test_commands(capsys):
    """Verify commands work"""
    call_command("pgtrigger", "ls")
    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.CustomSoftDelete:soft_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.CustomTableName:protect_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.FSM:fsm",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SearchModel:add_body_title_to_vector",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SearchModel:add_body_to_vector",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestDefaultThrough:protect_it",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTrigger:protect_misc_insert",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ToLogModel:after_update_row_test",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ToLogModel:after_update_statement_test",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ToLogModel:update_of_statement_test",
        "\x1b[94mUNALLOWED\x1b[0m \x1b[94mN/A\x1b[0m     tests.OrderSchema:protect_deletes",
        "\x1b[94mUNALLOWED\x1b[0m \x1b[94mN/A\x1b[0m     tests.ReceiptSchema:protect_updates",
    ]
    assert set(expected_lines).issubset(set(lines))

    call_command("pgtrigger", "ls", "-d", "receipt")
    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "",
        "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m     tests.ReceiptSchema:protect_updates",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.CustomSoftDelete:soft_delete",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.CustomTableName:protect_delete",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.FSM:fsm",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.SearchModel:add_body_title_to_vector",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.SearchModel:add_body_to_vector",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.TestDefaultThrough:protect_it",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.TestTrigger:protect_misc_insert",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.ToLogModel:after_update_row_test",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.ToLogModel:after_update_statement_test",
        "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.ToLogModel:update_of_statement_test",
        "\x1b[94mUNALLOWED\x1b[0m   \x1b[94mN/A\x1b[0m     tests.OrderSchema:protect_deletes",
    ]
    assert set(expected_lines).issubset(set(lines))

    call_command("pgtrigger", "install", "-d", "receipt")
    call_command("pgtrigger", "ls", "-d", "receipt")
    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ReceiptSchema:protect_updates",
    ]
    assert set(expected_lines).issubset(set(lines))

    # Installed a trigger to be pruned.
    protect_inserts = pgtrigger.Protect(name="protect_inserts", operation=pgtrigger.Insert)
    protect_inserts.install(models.OrderSchema, database="order")

    call_command("pgtrigger", "ls", "-d", "order")
    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m     tests.OrderSchema:protect_deletes",
        "\x1b[94mUNALLOWED\x1b[0m   \x1b[94mN/A\x1b[0m     tests.ReceiptSchema:protect_updates",
        "\x1b[96mPRUNE\x1b[0m       \x1b[92mENABLED\x1b[0m tests_orderschema:pgtrigger_protect_inserts_a0767",
    ]
    assert set(expected_lines).issubset(set(lines))

    call_command("pgtrigger", "prune", "-d", "order")
    call_command("pgtrigger", "install", "-d", "order")
    call_command("pgtrigger", "ls", "-d", "order")
    lines = capsys.readouterr().out.split("\n")
    for line in lines:
        assert "PRUNE" not in line
    expected_lines = [
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.OrderSchema:protect_deletes",
        "\x1b[94mUNALLOWED\x1b[0m \x1b[94mN/A\x1b[0m     tests.ReceiptSchema:protect_updates",
    ]
    assert set(expected_lines).issubset(set(lines))

    # Set the search path to a schema and check results
    call_command("pgtrigger", "uninstall", "-s", "receipt")
    call_command("pgtrigger", "ls", "tests.CustomSoftDelete:soft_delete", "-s", "receipt")
    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "",
        "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m tests.CustomSoftDelete:soft_delete",
    ]
    assert set(expected_lines) == set(lines)

    call_command("pgtrigger", "install", "tests.CustomSoftDelete:soft_delete", "-s", "receipt")
    call_command("pgtrigger", "ls", "tests.CustomSoftDelete:soft_delete", "-s", "receipt")
    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.CustomSoftDelete:soft_delete",
    ]
    assert set(expected_lines) == set(lines)



import pytest
from django.db import connection

import pgtrigger.tests.syncdb_app.models as syncdb_models
from pgtrigger.tests import utils


@pytest.mark.django_db
def test_create_model_creates_triggers():
    """
    Tests trigger installation with syncdb

    `DatabaseSchemaEditorMixin.create_model` is called when the django app doesn't
    have a migrations module. `DatabaseSchemaEditorMixin.create_model` is also called
    during a `CreateTable` migration operation but as the triggers aren't stored with
    the `CreateTable operation`, the specific code that creates triggers in
    `DatabaseSchemaEditorMixin.create_model` isn't executed.
    """
    with connection.schema_editor() as editor:
        editor.create_model(syncdb_models.NoMigrationModel)

    with utils.raises_trigger_error(match="no no no!"):
        syncdb_models.NoMigrationModel.objects.create(field="misc_insert", int_field=1)
